package devmgr.api.melutil;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

/**
This class provides a common set of methods for Resource classes to use.  Each package with a 
property file will automatically have a Resource.java file created from a template.  The Resource 
class will inherit from this BaseResource class.   
<p>
BaseResource provides a locking mechanism to ensure that only strings from a specific locale are 
retrieved as long as the lock is set. This can be useful for guaranteeing that support data, for 
example, is always in a specific locale. To minimize the impact of this locking mechanism, it should
do nothing if the user's system is running with the language set to the support locale or a language
we do not support.
<P>
All calls to getString, getCharacter and getIcon will return the global resource with the key, if  
the key is not found in the specified bundle. Therefore, there are no methods to getGlobalString or 
getGlobalIcon. 
<p>
An icon cache is used to keep icons from being read from disk each time they are accessed.  The icon
cache is a SoftHashMap.  The purpose of the SoftHashMap is to hold SoftReferences to the values.  A
SoftReference is a Reference object that can be garbage collected when the JVM is starting to run 
low on memory.  SoftRefences are garbage collected to gain the needed memory, starting from the 
least recently used.
<p>

NOTES/ASSUMPTIONS: <BR>
None.
*/
class BaseResource
{
   /**
    * This constant can be used to make sure the base properties file is accessed instead of a
    * locale-specific one.
    */
   public static final Locale     NO_LOCALE                 = new Locale("", "", "");

   /**
    * This is returned on the getChar method when we are unable to determine the value.
    */
   public static final char       UNKNOWN_CHAR              = '\0';

   protected static final String  GLOBAL_RESOURCE_NAME      = "devmgr/versioned/shared/SharedStrings";
  
   private static Map<Thread, Locale> m_LocaleLockMap;
   private static boolean             m_IsForcingLocale;

   
   /**
   Will return the fully qualified path for the given Class and Relative Path.  If the Relative
   Path is an empty string (""), then this method will return that empty string.
   <P>
   NOTES/ASSUMPTIONS:  <BR>
   None.
   
   @param theClass Class of the Resource Class trying to load an image icons
   @param relativeImagePath Relative path location for the images
   @return Fully qualified path for the passed in class.
   */
   protected static String getImagePath(Class theClass, String relativeImagePath)
   {
      String            addedPath;
      String            classPath;
      int               lastIndex = 0;
      int               findCount = 0;
      int               stringLocation;
      String            finalImagePath = relativeImagePath;
      
      if (finalImagePath.length() > 0)
      {
         addedPath = finalImagePath.replaceAll("\\.\\.\\/", "");
         classPath = theClass.getPackage().getName(); 
         
         while ((lastIndex = finalImagePath.indexOf("../", lastIndex)) != -1)
         {
            // Increment the lastIndex so we don't find the same one again
            lastIndex += 1;
            findCount += 1;
         }
         
         for (int i=0; i < findCount; i++)
         {
            stringLocation = classPath.lastIndexOf('.');
            if (stringLocation == -1)
            {
               classPath = "";
            }
            else
            {
               classPath = classPath.substring(0, stringLocation);
            }
         }
   
         finalImagePath = "";
         if (classPath.length() > 0)
         {
            finalImagePath += '/' + classPath.replace('.', '/');
         }

         finalImagePath += '/' + addedPath;
      }
      
      return finalImagePath;
   }
   
   /**
    BaseResource must not be instantiated.
    <P>
    NOTES/ASSUMPTIONS:  <BR>
    None.
    */
   protected BaseResource()
   {
      //Intentionally left blank.
   }
   
   /**
   This method should be called to guarantee that all strings retrieved from resource bundles in 
   the current thread are in English until unlock is called. After being locked, calls to getString 
   from a thread other than the locking thread will proceed as normal with the default or specified 
   locale.
   <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   Care should be taken to guarantee that the thread that does the locking eventually invokes
   unlock; otherwise, the thread will continue to use English without restoring the default locale. 
   <P>
   If the system's language is either English or an unsupported language, the base properties files
   are used, which are in English, so this method does nothing in these cases.
   
   @see #lockForLocale
   @deprecated Use lockForLocale(Locale).
   */
   public static void lockForEnglish()
   {
      lockForLocale(Locale.ENGLISH);
   }
   
   /**
   Indicates whether a non-English language is supported. <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   None.
   
   @return a boolean indicated whether the non-English language is supported
   */
   public static boolean isLocaleLanguageSupported()
   {
	   //TODO: fix
      return true;
   }
   
   /**
   This method should be called to guarantee that all strings retrieved from resource bundles in 
   the current thread are in the specified locale until unlock is called. After being locked, calls 
   to getString from a thread other than the locking thread will proceed as normal with the default 
   or specified locale.
   <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   Care should be taken to guarantee that the thread that does the locking eventually invokes
   unlock; otherwise, the thread will continue to use the specified locale without restoring the 
   default locale. 
   <P>
   If the system's language is either English or an unsupported language, the base properties files
   are used, which are in English, so this method does nothing in these cases.
   
   @param locale The locale to lock for the current thread. 
   
   @see #unlock
   */
   public static synchronized void lockForLocale(Locale locale)
   {
      // m_LocaleLanguageSupported will only be true if the user's system is set to run with
      // one of the non-English languages we support. And only then do we ever have to worry about
      // the locking mechanism, since all strings retrieved would be English anyway in all
      // other scenarios.
      if (locale != Locale.getDefault())
      {
         if (m_LocaleLockMap == null)
         {
            m_LocaleLockMap = new HashMap<Thread, Locale>();
         }
         // We don't check to see if the thread already exists in the HashMap, because the put does
         // the check and won't do anything if it already exists.
         m_LocaleLockMap.put(Thread.currentThread(), locale);
         m_IsForcingLocale = true;
      }
      
   }
   
   /**
   Clears the lock that guarantees only the specified locale Strings are being retrieved. <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   None.
   
   @see #lockForLocale
   */
   public static synchronized void unlock()
   {
      // Since the locale lock is only being used when a non-default locale needs to be forced, we 
      // only need to do something here if m_ForcingLocale was set to true by an invocation of 
      // lockForLocale().
      if (m_IsForcingLocale)
      {
         if (m_LocaleLockMap.remove(Thread.currentThread()) == null)
         {
            // TODO: trace
         }
         else if (m_LocaleLockMap.isEmpty())
         {
            m_IsForcingLocale = false;
            m_LocaleLockMap = null;
         }
      }
   }
   
   /**
   Indicates whether the resource bundles are being forced to return English locale data. <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   None.
   
   @return a boolean indicated whether the resource bundle has been locked to force English only
   @deprecated use isForcingLocale() or isForcingLocale(Locale).
   */
   public static boolean isForcingEnglish()
   {
      return isForcingLocale(Locale.ENGLISH);
   }
   
   /**
    Indicates whether the resource bundles for the current thread are being forced to a non-default
    locale data.
    <P>
    
    NOTES/ASSUMPTIONS: <BR>
    None.
    
    @return Indicates whether the resource bundle has been locked to force a non-default locale.
    */
   public static boolean isForcingLocale()
   {
      return m_IsForcingLocale && m_LocaleLockMap.containsKey(Thread.currentThread());
   }
   
   /**
    Indicates whether the resource bundles for the current thread are being forced to return a
    specific locale data.
    <P>
    
    NOTES/ASSUMPTIONS: <BR>
    None.
   
    @param locale The locale to check.
   
    @return a boolean indicated whether the resource bundle has been locked to force a specific
            locale.
   */
   public static boolean isForcingLocale(Locale locale)
   {
      return m_IsForcingLocale && m_LocaleLockMap.get(Thread.currentThread()) == locale;
   }
   
   /**
   Always returns an English string for the given keyString. <P>
   
   NOTES/ASSUMPTIONS:  <BR>
   This method should be called after waiting for the locale lock.
   
   @param keyString The key to the string in the resource
   @param baseResourceName The base resource name, including the path
   @return The String that matches the key or an empty string if the key is not found
   @deprecated use lockForEnglish or lockForLocale and the call getString.  The strings returned 
      will be all in the locale used in the lock.
   */
   protected static String getEnglishString(String keyString, String baseResourceName)
   {
      return getString(keyString, baseResourceName, Locale.ENGLISH, GLOBAL_RESOURCE_NAME, true);
   }

   /**
    * Get a string from the resource bundle based on the key. <BR>
    * <BR>
    * 
    * Notes/Assumptions: <BR>
    * If the key is not found in the bundle indicated by the baseResourceName, the global resources
    * are checked to find a global resource with the keyString.<br>
    * Always returns a string. Never returns a null. If the key is not found in the list, an empty
    * string is returned.
    * 
    * @param keyString The text identifier to the message in the resource bundle
    * @param baseResourceName The base resource name, including the path.
    * @return String that matches the key or an empty string.
    * 
    */
   protected static String getString(String keyString, String baseResourceName)
   {
      return getString(keyString, baseResourceName, getLocale(), GLOBAL_RESOURCE_NAME, true);
   }
   
   /**
    * Get a string from the resource bundle based on the key.
    * <P>
    * 
    * Notes/Assumptions: <BR>
    * If the key is not found in the bundle indicated by the baseResourceName, the global resources
    * are checked to find a global resource with the keyString.<br>
    * Always returns a string. Never returns a null. If the key is not found in the list, an empty
    * string is returned.
    * 
    * @param keyString The text identifier to the message in the resource bundle
    * @param baseResourceName The base resource name, including the path.
    * @param allowTracing Indicates if tracing should be allowed.
    * @return String that matches the key or an empty string.
    * 
    */
   protected static String getString(String keyString, String baseResourceName,
      boolean allowTracing)
   {
      return getString(keyString, baseResourceName, getLocale(), GLOBAL_RESOURCE_NAME,
         allowTracing);
   }

   /**
    * Get the locale for the current thread. <BR>
    * <BR>
    * 
    * Notes/Assumptions: <BR>
    * None.
    * 
    * @return The locale for this thread.
    */
   protected static Locale getLocale()
   {
      Locale locale = null;
      
      if (m_IsForcingLocale)
      {
         locale = m_LocaleLockMap.get(Thread.currentThread());
         if (locale == null)
         {
            locale = Locale.getDefault();
         }
      }
      else
      {
         locale = Locale.getDefault();
      }
      
      return locale;
   }

   /**
    * Get a string from the resource bundle based on the key. <BR>
    * <BR>
    * 
    * Notes/Assumptions: <BR>
    * Always returns a string. Never returns a null. If the key is not found in the list, an empty
    * string is returned.
    * 
    * @param keyString The text identifier to the message in the resource bundle
    * @param baseResourceName The base resource name, including the path.
    * @param locale The locale for the resource bundle.
    * @param globalResourceName The name of the global resource bundle.
    * @param allowTracing Indicates if tracing should be allowed. This is used because getIcon etc.
    *           handle their own tracing and using false avoids duplicate messages.
    * @return String that matches the key or an empty string.
    * 
    */
   protected static String getString(String keyString, String baseResourceName, Locale locale,
      String globalResourceName, boolean allowTracing)
   {
      ResourceBundle bundle;
      String outString = "";

      if (baseResourceName == null)
      {
         // TODO: trace
//    	  System.out.println(
//            "BaseResource.getString - A null resource bundle name.");
         return outString;
      }

      if (locale == null)
      {
         //TODO: trace
//    	  System.out.println("BaseResource.getString - A null locale.");
         return outString;
      }

      try
      {
         bundle = ResourceBundle.getBundle(baseResourceName, locale);
         outString = bundle.getString(keyString);
      }
      catch (MissingResourceException e)
      {
         // If the caller allows global resources to be used and the bundle being used isn't the
         // global resources, then try getting the string from the global resources.
         //System.out.println("Exception " + e.getLocalizedMessage());
      }
      catch (Exception e)
      {
         //System.out.println("Exception " + e.getLocalizedMessage());
      }

      return outString;
   }
   
   /**
    * Returns a string from the global resource bundle that contains generic strings to be used in
    * all Sam modules to provide a consistent user interface. <BR>
    * <BR>
    * 
    * @param keyString The key to the string in the resource
    * @return The string that matches the key or an empty string.
    * @deprecated Use getString(String).  getString will return global strings.
    */
   public static String getGlobalString(String keyString)
   {
      return getString(keyString, GLOBAL_RESOURCE_NAME, getLocale(), null, true);
   }
       
   /**
    * Get a char from the resource bundle based on the key. This method was added to support getting
    * a char to pass to the setMnemonic() methods. <BR>
    * <BR>
    * 
    * Notes/Assumptions: <BR>
    * If the key is not found in the bundle indicated by the baseResourceName, the global resources
    * are checked to find a global resource with the keyString.<br>
    * Always returns a char, but the char will be equal to UNKNOWN_CHAR if the key is not found in 
    * the list.
    * 
    * @param keyString The key to the character in the resource bundle
    * @param baseResourceName The base resource name, including the path.
    * @return char that matches the key or the UNKNOWN_CHAR.
    */
   protected static char getCharacter(String keyString, String baseResourceName)
   {
      return getCharacter(keyString, baseResourceName, GLOBAL_RESOURCE_NAME);
   }
   
   /**
    * Get a char from the resource bundle based on the key. This method was added to support getting
    * a char to pass to the setMnemonic() methods. <BR>
    * <BR>
    * 
    * Notes/Assumptions: <BR>
    * If the key is not found in the bundle indicated by the baseResourceName, the global resources
    * are checked to find a global resource with the keyString.<br>
    * Always returns a char, but the char will be equal to UNKNOWN_CHAR if the key is not found in 
    * the list.
    * 
    * @param keyString The key to the character in the resource bundle
    * @param baseResourceName The base resource name, including the path.
    * @param globalResourceName The name of the global resource bundle.
    * @return char that matches the key or the UNKNOWN_CHAR.
    */
   protected static char getCharacter(String keyString, String baseResourceName, 
      String globalResourceName)
   {
      String outString = getString(keyString, baseResourceName, getLocale(), globalResourceName,
         true);
      return outString.length() > 0 ? outString.charAt(0) : UNKNOWN_CHAR;
   }
   
   /**
    * Get a char from the global resource bundle based on the key. This method was added to support
    * getting a char to pass to the setMnemonic() methods.
    * <P>
    * 
    * NOTES/ASSUMPTIONS: <BR>
    * Calls getChar() with a reference to the global resource. Always returns a char, but the char
    * will be equal to null if the key is not found in the list.
    * <P>
    * 
    * @param keyString The key to the character in the resource
    * @return char that matches the key or the null character.
    * @deprecated Use Resource.getCharacter(String) instead. getCharacter will return global
    *             characters.
    */
   public static char getGlobalChar(String keyString)
   {
      return getCharacter(keyString, GLOBAL_RESOURCE_NAME, null);
   }

}
