package devmgr.api.melutil;

import devmgr.api.symbol.EventPriority;
import devmgr.api.symbol.MelEntry;
import devmgr.api.symbol.TraySlot;

import java.util.Date;
import java.util.Locale;

/**

DESCRIPTION: <BR>
The EventDecoder is provided to "decode" the various fields contained in the 
melEntry to string format for use by the event viewer.
<BR><BR>

NOTES/ASSUMPTIONS: <BR>
A single instance of EventDecoder will be created by the EventViewerJDialog
to handle the decoding of event records for the summary table display, the
save as function and the detail display.
<BR><BR>

REFERENCE: <BR>
User Interface Design - Event Viewer (349-1001830)
Design Detail - Event Viewer (349-1002090)
Feature Approach Documnet - Event Viewer (349-1001540)
*/
public class EventDecoder
{  

   /**
   DESCRIPTION: <BR>
   The event decoder will create an instance if the detail manager.  Detail 
   manager is an inner class of EventDecoder and is responsible for formatting
   and displaying event details.
   */
   
   private static final int         MILLISECOND_MULTIPLIER = 1000;
   static final String              m_NewLine = System.getProperty("line.separator");
   
   
   
   /**
   DESCRIPTION: <BR>
   Default constructor. 
   <BR> <BR>

   NOTES/ASSUMPTION: <BR>
   None.
   <BR><BR>
    */
   public EventDecoder()
   {
	   
   }

   public static void lockBundle(Locale locale)
   {
	   Resource.lockForLocale(locale);
	   EventUtil.init();
   }

   public static void unlockBundle()
   {
	   Resource.unlock();
   }
 

   /**
   DESCRIPTION: <BR>
   Converts the date as returned by the controller into an EventDate object
   representing the date of the event. The formatting is done by the toString()
   method of the EventDate class.
   <BR><BR>

   @param event      The event to request the date from.
   @return Date object representation of the event date.
   */
   public static Date getDate(final MelEntry event)
   {
      final Date date = new Date();
      //Time in seconds
      final long timeStamp = event.getTimeStamp();
      // This takes a long and wants it in milliseconds.
      date.setTime(timeStamp * MILLISECOND_MULTIPLIER);
      return date;
   }

   /**
   DESCRIPTION: <BR>
   Converts an integer value to the corresponding string representation of the
   event priority.
   <BR><BR>

   @param event      The event to request the priority from.
   @return String       String representation of the event priority.
   */
   public static EventPriority getEventPriority(MelEntry event)
   {
      return EventUtil.getEventPriority(event);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which converts an integer value to the 
   corresponding string representation of the component type.
   <BR> <BR>

   @param event    The event to request the component type from.
   @return String     String representation of the component type.
   */
   public static String getComponentType(MelEntry event)
   {
      return EventUtil.getComponentType(event);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which maps the event type to a string
   description of the event.
   <BR> <BR>

   @param event    The event to request the description from.
   @return String     Textual description of the event.
   */
   public static String getDescription(MelEntry event)
   {
      return EventUtil.getDescription(event);
   }
   
   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which maps the event type to a string
   description of the event and appends details of the event to it.
   <BR> <BR>

   @param event    The event to request the description from.
   @return String     Textual description of the event.
   */
   public static String getTotalDescription(MelEntry event)
   {
      return EventUtil.getTotalDescription(event);
   }
   
   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which maps the event type to a US ASCII 
   string description of the event.
   NOTE: java.util.Locale.US is a constant that creates a Locale object for _en_US. 
   The code will search for devmgr.versioned.sam.eventviewer.ResourceStrings_en_US and if it
   does not find this it will search for devmgr.versioned.sam.eventviewer.ResourceStrings_en
   and if it does not find this it will search for devmgr.versioned.sam.eventviewer.ResourceStrings.
   This will allow the default to be US, but also allow different defaults if the need arises 
   in the future.
   <BR> <BR>

   @param event    The event to request the description from.
   @return String     Textual description of the event.
   */
   public static String getUSASCIIDescription(MelEntry event)
   {
      return EventUtil.getUSASCIIDescription(event);
   }

   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which takes the sense key, ASC, and ASCQ fields
   in the melEntry and converts them to a string of the form XX/YY/ZZ. Where XX represents the 
   sense key, YY represents the ASC and ZZ represents the ASCQ.
   <BR> <BR>

   @param event    The event to request the ASC/ASCQ from.
   @returns String     Sense Key/ASC/ASCQ
   */
   public static String getAscAscq(MelEntry event)
   {
      return EventUtil.getAscAscq(event);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which converts the mel component location to a 
   string representation of the component location.  returns "Not applicable" if location valid
   in the melEntry is set to
   false.
   <BR> <BR>

   @param event    The event to request the component location from.
   @return String     Location of the affected component.
    */
   public static String getComponentLocation(MelEntry event)
   {
      return EventUtil.getComponentLocation(event);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which returns the appropriate string -
   Tray X Slot Y.
   <BR><BR>

   @param traySlot        Tray and slot number to decode.
   @return String    String to display in the detail and saved views.
   */
   public static String decodeTraySlot(TraySlot traySlot)
   {
      return EventUtil.decodeTraySlot(traySlot);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which checks for -1 in tray field and returns
   the appropriate string.
   <BR><BR>

   @param tray        Tray number to decode.
   @return String    String to display in the detail and saved views.
   */
   public static String decodeTray(int tray)
   {
      return EventUtil.decodeTray(tray);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which converts the 64 bit sequence number
   returned in the melEntry to a string.
   <BR> <BR>

   @param event    The event to request the sequence number from.
   @return String     Sequence number of the event.
   */
   public static String getSequenceNumber(MelEntry event)
   {
      return Long.toString(event.getSequenceNumber());
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which converts the event type contained in 
   the melEntry into a string.

   @param event    The event to request the event type from.
   @return String     Event type as a 4 digit string(hex).  
   */
   public static String getEventType(MelEntry event)
   {
      return EventUtil.getEventType(event);
   }


   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which converts an integer value to the
   corresponding string representation of the
   event category.

   @param event    The event to request the category from.
   @return String String representation of the event category.
   */
   public static String getEventCategory(MelEntry event)
   {
      return EventUtil.getEventCategory(event);
   }
   /**
   DESCRIPTION: <BR>
   Calls public static method in EventUtil.java which converts an integer value to the
   corresponding string representation of the
   event category.

   @param event    The event to request the category from.
   @return String String representation of the event category.
   */
   public static String getEventSourceController(MelEntry event)
   {
      return  EventUtil.getEventSourceController(event);       
   }
 }

