/*******************************************************************************
 * Copyright (c) 2016-2017 of NetApp, Inc. All Rights Reserved.                *
 *                                                                             *
 ******************************************************************************/
package devmgr.api.model.ldap;

import io.swagger.annotations.ApiModel;
import io.swagger.annotations.ApiModelProperty;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

/**
 * Model used to store data about an LDAP domain.
 *
 * @author vtroy
 */
@ApiModel(description = "Structure that contains information about an LDAP Domain")
public class LdapDomain implements Serializable {
    private static final long serialVersionUID = -323706224043976665L;

    @ApiModelProperty(value = "Unique ID linked to this domain")
    private String id;

    @ApiModelProperty(
          value = "If needed, you can configure a specific user to use when looking "
                + "up the group membership for users. Typically, you would configure "
                + "this whenever regular users might lack reader permissions to view "
                + "their own group membership. The bind user information must be specified as a full DN.")
    private BindLookupUser bindLookupUser;

    @ApiModelProperty(
          required = true,
          value = "A list of group attributes on the user that will be searched to for group to role mapping",
          example = "memberOf")
    private List<String> groupAttributes;

    @ApiModelProperty(
          required = true,
          value = "The LDAP URL entry must be specified as either ldap or ldaps protocol "
                + "and contain the IP address. In addition, the port for the LDAP URL "
                + "must be specified (typically 389 for ldap and 636 for ldaps).",
          example = "ldaps://ldap.example.com:636")
    private String ldapUrl;

    @ApiModelProperty(required = true, value = "Each domain will have one or"
          + "multiple names and it is presumed the name will match the DNS domain for the LDAP "
          + "server but it is not required. Domains can be named anything as long as "
          + "they are valid DNS names.")
    private List<String> names;

    @ApiModelProperty(
          value = "A list of regular expression patterns to match to the user's group "
                + "attributes to match to roles. (TODO: NEED TO FIX FOR A MAPPING TYPE)")
    private List<GroupMapping> roleMapCollection;

    @ApiModelProperty(
          value = "The search base is used to find group memberships of the user. "
                + "The search base is the DN in the directory of a container object of users. "
                + "The filter base is used to find the user object within this container. "
                + "After the user object is located, any associated group membership is identified.",
          example = "ou=users,dc=example,dc=com")
    private String searchBase;

    @ApiModelProperty(
          value = "The user attribute is used match the username supplied to an object on the server",
          example = "sAMAccountName")
    private String userAttribute;

    public LdapDomain() {

    }

    public final BindLookupUser getBindLookupUser() {
        return bindLookupUser;
    }

    public final List<String> getGroupAttributes() {
        if(groupAttributes == null) {
            return new ArrayList<>();
        }
        return groupAttributes;
    }

    public final String getLdapUrl() {
        return ldapUrl;
    }

    public final List<String> getNames() {
        if(names == null) {
            return new ArrayList<>();
        }
        return names;
    }

    public final List<GroupMapping> getRoleMapCollection() {
        if(roleMapCollection == null) {
            return new ArrayList<>();
        }
        return roleMapCollection;
    }

    public final String getSearchBase() {
        return searchBase;
    }

    public final String getUserAttribute() {
        return userAttribute;
    }

    public String getId() {
        return id;
    }

    public final void setBindLookupUser(BindLookupUser bindLookupUser) {
        this.bindLookupUser = bindLookupUser;
    }

    public final void setGroupAttributes(List<String> groupAttributes) {
        this.groupAttributes = groupAttributes;
    }

    public final void setLdapUrl(String ldapUrl) {
        this.ldapUrl = ldapUrl;
    }

    public final void setNames(List<String> names) {
        this.names = names;
    }

    public final void setRoleMapCollection(List<GroupMapping> rmc2) {
        this.roleMapCollection = rmc2;
    }

    public final void setSearchBase(String searchBase) {
        this.searchBase = searchBase;
    }

    public final void setUserAttribute(String userAttribute) {
        this.userAttribute = userAttribute;
    }

    public void setId(String id) {
        this.id = id;
    }

    /**
     * Sets the ID to a unique string
     */
    public void generateAndSetID() {
        if(id == null) {
            id = UUID.randomUUID().toString();
        }
    }

    @Override
    public String toString() {
        return "LdapDomain [groupAttributes=" + groupAttributes + ", ldapUrl=" + ldapUrl + ", name=" + names +
              ", roleMapCollection=" + roleMapCollection + ", searchBase=" + searchBase + "]";
    }

    @Override
    public boolean equals(Object o) {
        if(this == o) {
            return true;
        }
        if(!(o instanceof LdapDomain)) {
            return false;
        }
        LdapDomain domain = (LdapDomain) o;

        if(domain.getBindLookupUser() != null) {
            if(!domain.getBindLookupUser().equals(bindLookupUser)) {
                return false;
            }
        }

        if(domain.getGroupAttributes().size() != getGroupAttributes().size()) {
            return false;
        }

        for(int i = 0, size = domain.getGroupAttributes().size(); i < size; i++) {
            if(!domain.getGroupAttributes().get(i).equals(getGroupAttributes().get(i))) {
                return false;
            }
        }

        if(domain.getNames().size() != getNames().size()) {
            return false;
        }

        for(int i = 0, size = domain.getNames().size(); i < size; i++) {
            if(!domain.getNames().get(i).equals(getNames().get(i))) {
                return false;
            }
        }

        if(domain.getRoleMapCollection().size() != getRoleMapCollection().size()) {
            return false;
        }

        for(int i = 0, size = domain.getRoleMapCollection().size(); i < size; i++) {
            if(!domain.getRoleMapCollection().get(i).equals(getRoleMapCollection().get(i))) {
                return false;
            }
        }

        boolean searchBaseMatches = domain.getSearchBase() == null ?
                searchBase == null : domain.getSearchBase().equals(searchBase);
        boolean userAttributeMatches = domain.getUserAttribute() == null ?
                userAttribute == null : domain.getUserAttribute().equals(userAttribute);
        boolean idMatches = domain.getId() == null ? id == null : domain.getId().equals(id);
        boolean urlMatches = domain.getLdapUrl() == null ? ldapUrl == null : domain.getLdapUrl().equals(ldapUrl);
        if(!searchBaseMatches || !userAttributeMatches || !idMatches || !urlMatches) {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        int idHash = id != null ? id.hashCode() : 0;
        int bindUserHash = bindLookupUser != null ? bindLookupUser.hashCode() : 0;
        int groupAttributesHash = groupAttributes != null ? groupAttributes.hashCode() : 0;
        int urlHash = ldapUrl != null ? ldapUrl.hashCode() : 0;
        int namesHash = names != null ? names.hashCode() : 0;
        int roleMapHash = roleMapCollection != null ? roleMapCollection.hashCode() : 0;
        int searchBaseHash = searchBase != null ? searchBase.hashCode() : 0;
        int userAttributeHash = userAttribute != null ? userAttribute.hashCode() : 0;

        return 191 * (idHash + bindUserHash + groupAttributesHash + urlHash + namesHash + roleMapHash + searchBaseHash +
                userAttributeHash);
    }
}
