package devmgr.api.rpc;

import devmgr.api.symbol.SYMbolAPI;
import java.io.IOException;
import java.net.InetAddress;
import java.util.concurrent.TimeoutException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public interface Client extends AbstractRPCControllerClient, AutoCloseable {

    public static final SYMbolAPI api = new SYMbolAPI();

    public abstract int getXid();

    public abstract void setControllerRef(String controllerRef);

    public Throwable getLastError();

    /**
     * Close the connection.
     *
     * @throws IOException if there's an IO error
     */
    public void close() throws IOException;

    /**
     * Connect to the server synchronously (blocking).
     *
     * @param server the remote host
     * @throws Exception if there's an error
     */
    public abstract void connect(InetAddress server) throws Exception;

    /**
     * Initiate the server connection synchronously (blocking).
     *
     * @param server the remote host
     * @param port the port on the remote host
     * @throws Exception if there's an error
     */
    public abstract void connect(InetAddress server, int port) throws Exception;

    /**
     * Connect to the remote host asynchronously (non-blocking).
     *
     * @param server the remote host
     * @param callback the callback used to communicate success or failure
     */
    public abstract void connect(InetAddress server, NIOClientCallback callback);

    /**
     * Initiate the server connection asynchronously (non-blocking).
     *
     * @param server the remote host
     * @param port the port on the remote host
     * @param callback the callback used to communicate success or failure
     */
    public abstract void connect(InetAddress server, int port, NIOClientCallback callback);

    /**
     * Check if the connection is open.
     *
     * @return true if connected
     */
    public abstract boolean isConnected();

    /**
     * Return the IP address of the client end of the connection.
     *
     * @return the client-side ip address
     */
    public abstract InetAddress getLocalAddress();

    /**
     * Return the remote host's address.
     *
     * @return the remote host inetaddress
     */
    public abstract InetAddress getDestinationIP();


    // This class is used if the user does not specify a callback, making the calls a blocking call
    static class BlockingCallbackHandler<T> implements NIOClientCallback<T> {

        private static final Logger logger = LoggerFactory.getLogger(BlockingCallbackHandler.class);
        Exception lastError;
        T lastResults;
        volatile boolean done;

        @Override
        public void onTimeout() {
            logger.debug("Blocking handle had a timeout");
            lastError = new TimeoutException("timeout");
            signal();
        }

        boolean isDone() {
            return done;
        }

        private void signal() {
            done = true;
            synchronized (this) {
                notifyAll();
            }
        }

        @Override
        public void onError(Exception t) {
            logger.debug("Blocking handle had an error");
            lastError = t;
            signal();
        }

        @Override
        public void onSuccess(T results) {
            logger.debug("Blocking handle was successful");
            lastResults = results;
            signal();
        }
    }
}
