package devmgr.api.rpc;

import devmgr.api.symbol.SYMbolAPI;
import java.io.IOException;
import java.net.InetAddress;
import java.nio.channels.SocketChannel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This is used to build a concrete implementation of abstract class {@link Client}.
 */
public class ClientTransportBuilder {

    private static final Logger logger = LoggerFactory.getLogger(ClientTransportBuilder.class);
    private SelectManager selectManager;
    private SocketChannel socketChannel;
    private InetAddress inetAddress;
    private int rpcPort;
    private int httpPort;
    private boolean enableHttpTunnel;

    /**
     * @param selectManager the SelectKey channel manager
     * @param inetAddress the remote host
     */
    public ClientTransportBuilder(SelectManager selectManager, InetAddress inetAddress) {
        this(selectManager, inetAddress, null);
    }

    public ClientTransportBuilder(SelectManager selectManager, InetAddress inetAddress,
            SocketChannel socketChannel) {
        this.selectManager = selectManager;
        this.inetAddress = inetAddress;
        this.socketChannel = socketChannel;
        this.rpcPort = SYMbolAPI.PORT;
        this.httpPort = HttpRpcClient.PORT;
        // The property is defined only in SMclient code.
        this.enableHttpTunnel = Boolean.getBoolean("UseHttpAllowed");
    }

    /**
     * Set a custom RPC port if the default {@link SYMbolAPI#PORT} is not applicable. The port is
     * needed since the builder logic needs to connect to the remote host.
     *
     * @param port the rpc port
     * @return an instance to the builder
     */
    public ClientTransportBuilder rpcPort(int port) {
        this.rpcPort = port;
        return this;
    }

    /**
     * Set a custom HTTP port if the default {@link HttpRpcClient#PORT} is not applicable.
     *
     * @param port the port
     * @return an instance to the builder
     */
    public ClientTransportBuilder httpPort(int port) {
        this.httpPort = port;
        return this;
    }

    /**
     * Set this to true to to try using the http tunnel as the rpc transport.
     *
     * @param enableHttpTunnel try using the http to carry the rpc bits
     * @return an instance to the builder
     */
    public ClientTransportBuilder enableHttpTunnel(boolean enableHttpTunnel) {
        this.enableHttpTunnel = enableHttpTunnel;
        return this;
    }

    /**
     * Build the object.
     *
     * @return a child instance of the abstract class
     * @throws RPCError when there's an RPC error
     * @throws IOException when there's an IO error
     */
    public Client build() throws RPCError, IOException {
        if (!enableHttpTunnel) {
            logger.debug("Defaulting to using symbol tcp socket for address {}", inetAddress);
            return new NIORPCClient(selectManager, socketChannel);
        }

        // TODO(blucas): The code below may be SLOW and BLOCK. This will need to be solved
        // in the future. Ignoring for now since this runs in the context of the EMW which is slow.
        // Likely easier to find a solution high up the stack.
        try (Client client = new NIORPCClient(selectManager, socketChannel)) {
            logger.debug("Attempting to use symbol tcp socket for address {}", inetAddress);
            client.connect(inetAddress, rpcPort);
            client.call("pingController", null);
            // The connection and ping are successful. Use the symbol tcp transport.
            logger.debug("Choosing symbol tcp for transport");
            return new NIORPCClient(selectManager, socketChannel);
        } catch (Exception e) {
            // Do Nothing
            logger.debug("Failed to use symbol tcp transport, will try the http tunnel", e);
        }

        try {
            // Try a connection using the http transport. If the ctor returns we have a valid connection.
            Client connection = new HttpRpcClient(inetAddress, httpPort,
                    HttpRpcClient.SCHEME_HTTPS);
            logger.debug("Choosing http tunnel for symbol transport");
            return connection;
        } catch (Exception e) {
            // Okay, so both transports failed (symbol tcp and http tunnel). Nothing we can do.
            logger.warn("No viable transport protocols found for {}", inetAddress);
            throw new IOException("No viable transport protocols", e);
        }
    }
}
