package devmgr.api.rpc;

import devmgr.api.rpc.RPCError.ErrorType;
import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.DecodeException;
import devmgr.api.runtime.Decoder;
import devmgr.api.runtime.EncodeException;
import devmgr.api.runtime.XdrMessage;
import devmgr.api.symbol.ReturnCode;
import devmgr.api.symbol.SYMbolAPI;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.net.InetAddress;
import java.net.URI;
import java.net.UnknownHostException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.OptionalInt;
import java.util.concurrent.Future;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;
import javax.ws.rs.ProcessingException;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.InvocationCallback;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriBuilder;
import javax.ws.rs.ext.MessageBodyReader;
import javax.ws.rs.ext.MessageBodyWriter;
import org.glassfish.jersey.client.ClientProperties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A {@link Client} class that uses the http tunnel to transport rpc packets.
 */
public class HttpRpcClient implements Client {

    private static final Logger LOG = LoggerFactory.getLogger(HttpRpcClient.class);
    private static final String RPC_TUNNEL_ENDPOINT = "devmgr/internal/symrpc";
    private static final String HTTP_NETAPP_CLIENT_TYPE_KEY = "netapp-client-type";
    private static final String HTTP_NETAPP_CLIENT_TYPE_VALUE = "rest-proxy";
    static final String SCHEME_HTTPS = "https";
    static final int PORT = 8443;

    private final XdrProtocolHelper protocolHelper = new XdrProtocolHelper(Client.api);
    private int xid = (int) ((new Date()).getTime() / 1000);
    private Throwable lastError;
    private String controllerRef;
    private OptionalInt userDefinedTimeout = OptionalInt.empty();
    private int timeout;
    private final InetAddress server;
    private final WebTarget target;

    HttpRpcClient(InetAddress server) throws KeyManagementException, NoSuchAlgorithmException {
        this(server, PORT, SCHEME_HTTPS);
    }

    HttpRpcClient(InetAddress server, int port, String scheme)
            throws NoSuchAlgorithmException, KeyManagementException {

        // TODO(blucas): The proxy doesn't need to verify the certs but perhaps an env prop
        // should be used to switch between allow all.
        SSLContext sslContext = SSLContext.getInstance("TLS");
        sslContext.init(null, new TrustManager[]{new X509TrustManager() {
            public void checkClientTrusted(X509Certificate[] arg0, String arg1)
                    throws CertificateException {
            }

            public void checkServerTrusted(X509Certificate[] arg0, String arg1)
                    throws CertificateException {
            }

            public X509Certificate[] getAcceptedIssuers() {
                return new X509Certificate[0];
            }

        }}, new java.security.SecureRandom());

        // Create the 'https://hostname:port/' uri
        URI uri = UriBuilder.fromUri("").host(server.getHostAddress()).scheme(scheme).port(port)
                .build();
        this.target = ClientBuilder.newBuilder().sslContext(sslContext).hostnameVerifier(
                new HostnameVerifier() {
                    @Override
                    public boolean verify(String s1, SSLSession s2) {
                        return true;
                    }
                }).register(XdrMessageBodyWriter.class).register(XdrMessageBodyReader.class)
                .build().target(uri);
        this.server = server;

        // Verify that the http tunnel is available.
        // NOTE: This is a BLOCKING call
        String res = target.path(RPC_TUNNEL_ENDPOINT)
                .request(MediaType.APPLICATION_JSON)
                .header(HTTP_NETAPP_CLIENT_TYPE_KEY, HTTP_NETAPP_CLIENT_TYPE_VALUE)
                .get(String.class);
        LOG.debug("Http tunnel is available: {}", res);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setTypeOverrides(Decoder.TypeOverrideTable typeOverrides) {
        this.protocolHelper.setTypeOverrides(typeOverrides);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <T> T call(String proc, Object args) throws Exception {
        ApiDefinition.ProcTableEntry x = Client.api.getProcByName(proc);
        return call(x.procNumber, args);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <T> T call(int proc, Object args) throws Exception {
        LOG.debug("Creating a BLOCKING handler for the call");
        final BlockingCallbackHandler callback = new BlockingCallbackHandler<T>();
        call(proc, args, callback);
        // NOTE: This synchronizes on a local variable intentionally. The `connect` call is using
        // the blocking handler which will notifyAll() when it is complete.
        synchronized (callback) {
            while (!callback.isDone()) {
                callback.wait(500);
            }
        }
        if (callback.lastError != null) {
            lastError = callback.lastError;
            throw callback.lastError;
        }
        return (T) callback.lastResults;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Object call(String proc, Object args, NIOClientCallback callback) {
        ApiDefinition.ProcTableEntry x = Client.api.getProcByName(proc);
        return call(x.procNumber, args, callback);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int call(final int procedureId, Object args, final NIOClientCallback callback) {
        this.xid = xid + 1;
        if (procedureId == 202) {
            LOG.debug("Setting success for validatePassword rpc (no network data sent)");
            callback.onSuccess(ReturnCode.ok);
            // TODO(blucas): Without a network transaction, should the xid change??
            return xid;
        }

        try {
            XdrMessage callMsg = protocolHelper.buildRequestMsg(procedureId, xid, args);
            LOG.debug("Tunnel: call for {} proc={} timeout={} xid={} to {}",
                    api.getProcByNumber(procedureId).procName, procedureId, timeout, xid,
                    target.getUri());

            // NOTE: Using `2*suggestedTimeout` comes from NIORPCClient.
            timeout = 2 * api.getProcByNumber(procedureId).suggestedTimeout;
            final Future<Response> entityFuture = target.path("devmgr/internal/symrpc")
                    .request()
                    .header(HTTP_NETAPP_CLIENT_TYPE_KEY, HTTP_NETAPP_CLIENT_TYPE_VALUE)
                    .property(ClientProperties.CONNECT_TIMEOUT, 100 * 1000)
                    .property(ClientProperties.READ_TIMEOUT,
                            1000 * userDefinedTimeout.orElse(timeout))
                    .async()
                    .post(Entity.entity(callMsg, MediaType.APPLICATION_OCTET_STREAM_TYPE),
                            new InvocationCallback<Response>() {
                                @Override
                                public void completed(Response response) {
                                    processResponse(procedureId, response, callback);
                                }

                                @Override
                                public void failed(Throwable throwable) {
                                    // NOTE: The failed(Throwable) method is only invoked because of
                                    // an internal client-side processing error
                                    LOG.error("Something bad happened procedureId={}", procedureId,
                                            throwable);
                                    callback.onError(
                                            new HttpRpcClientProcessingException(throwable));
                                }
                            });

        } catch (EncodeException | RPCError e) {
            LOG.error("Something bad happened procedureId={}", procedureId, e);
            callback.onError(e);
        }
        return xid;
    }

    /**
     * Process the response payload provided by the server. A 200 status indicates a successful
     * transaction and the entity is read from the stream. The Response's streams are closed
     * at the end of this call.
     *
     * @param procedureId the procId responsible for creating the corresponding response
     * @param s Response from the server
     * @param callback the callback used to communicate success or failure
     */
    private void processResponse(int procedureId, Response s, NIOClientCallback callback) {
        if (s.getStatus() != 200) {
            Exception e;
            switch (s.getStatus()) {
                case 400:
                case 415:
                case 500:
                case 502:
                    e = new RPCError(ErrorType.SYSTEM_ERROR);
                    break;

                case 401:
                    e = new RPCError(ErrorType.AUTH_REQUIRED);
                    break;

                case 403:
                    e = new RPCError(ErrorType.INVALID_RBAC_PERMS);
                    break;

                case 405:
                    e = new RPCError(ErrorType.PROC_UNAVAILABLE);
                    break;

                case 504:
                    e = new RPCError(ErrorType.TIMEOUT);
                    break;

                default:
                    e = new RPCError(ErrorType.SYSTEM_ERROR);
                    break;
            }
            LOG.info(
                    "Response status for procedureId={} is not okay: {}. Starting callback.onError",
                    procedureId,
                    s.getStatus());
            s.close();
            callback.onError(e);
        } else {
            LOG.debug(
                    "Okay http response. Processing reply message for procedureid={} and starting callback.onSuccess",
                    procedureId);
            try {
                // Process the reply message and notify the callback
                // NOTE: Jersey will close the streams when the entity is read. See Section 5.7 of
                // the manual https://jersey.java.net/documentation/latest/client.html#d0e5170
                XdrMessage message = s.readEntity(XdrMessage.class);
                Object result = protocolHelper.processReplyMsg(procedureId, message, this.xid);
                callback.onSuccess(result);
            } catch (RPCError | DecodeException e) {
                LOG.error("Something bad happened procedureId={}", procedureId, e);
                s.close();
                callback.onError(e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public int getXid() {
        return xid;
    }

    /**
     * {@inheritDoc}
     */
    public Throwable getLastError() {
        return lastError;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void close() throws IOException {
        LOG.debug("close() is a no-op for the http tunnel");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server) throws Exception {
        connect(server, SYMbolAPI.PORT);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server, int port) throws Exception {
        LOG.debug("Creating a BLOCKING handler for the connection to {}", server.getHostAddress());
        final BlockingCallbackHandler callback = new BlockingCallbackHandler();
        connect(server, port, callback);
        // NOTE: This synchronizes on a local variable intentionally. The `connect` call is using
        // the blocking handler which will notifyAll() when it is complete.
        synchronized (callback) {
            while (!callback.isDone()) {
                callback.wait(500);
            }
        }
        if (callback.lastError != null) {
            LOG.info("Blocking connect failed with error", callback.lastError);
            lastError = callback.lastError;
            throw callback.lastError;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server, NIOClientCallback callback) {
        connect(server, SYMbolAPI.PORT, callback);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server, int port, NIOClientCallback callback) {
        assert (server != null && this.server.getHostAddress().equals(server.getHostAddress()));
        callback.onSuccess(null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isConnected() {
        LOG.debug("http tunnel is assumed to be always open");
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public InetAddress getLocalAddress() {
        try {
            // TODO(blucas): I don't see a way to get the local connection from jersey. Getting the
            // local host can be weird if the system has many interfaces...
            return InetAddress.getLocalHost();
        } catch (UnknownHostException e) {
            LOG.error("Something bad happened", e);
            throw new RuntimeException("Failed get getLocalHost", e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public InetAddress getDestinationIP() {
        return server;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getControllerRef() {
        return controllerRef;
    }


    /**
     * {@inheritDoc}
     */
    @Override
    public void setControllerRef(String controllerRef) {
        this.controllerRef = controllerRef;
    }

    /**
     * {@inheritDoc}
     */
    public void setAuthGenerator(RPCAuthGenerator generator) {
        protocolHelper.setAuthGenerator(generator);
    }

    /**
     * Get the timeout in seconds.
     *
     * @return timeout of the command (seconds)
     */
    @Override
    public int getTimeout() {
        return userDefinedTimeout.orElse(timeout);
    }

    /**
     * Set a custom procedure timeout. Note that this is STICKY and re-use of an instance for
     * a different RPC call will use the previous timeout! This matches the timeout functionality of
     * NIORPCClient.
     *
     * @param timeoutInSeconds the custom procedure timeout (in seconds)
     */
    @Override
    public void setTimeout(int timeoutInSeconds) {
        userDefinedTimeout = OptionalInt.of(timeoutInSeconds);
    }


    public static class XdrMessageBodyReader
            implements MessageBodyReader<XdrMessage> {

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean isReadable(Class<?> type, Type genericType,
                java.lang.annotation.Annotation[] annotations,
                MediaType mediaType) {
            return type == XdrMessage.class;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public XdrMessage readFrom(Class<XdrMessage> type, Type genericType,
                java.lang.annotation.Annotation[] annotations,
                MediaType mediaType, MultivaluedMap<String, String> httpHeaders,
                InputStream entityStream) throws IOException, WebApplicationException {
            try {
                // NOTE: the channel should not close the stream. Jersey manages the stream.
                return new StreamingChannel(null, entityStream, false).receiveMsg();
            } catch (Exception e) {
                throw new ProcessingException("Error deserializing XdrMessage", e);
            }
        }
    }


    public static class XdrMessageBodyWriter
            implements MessageBodyWriter<XdrMessage> {

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean isWriteable(Class<?> type, Type genericType, Annotation[] annotations,
                MediaType mediaType) {
            return type == XdrMessage.class;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public long getSize(XdrMessage xdrMessage, Class<?> type, Type genericType,
                Annotation[] annotations, MediaType mediaType) {
            // deprecated by JAX-RS 2.0 and ignored by Jersey runtime
            return 0;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void writeTo(XdrMessage xdrMessage, Class<?> type, Type genericType,
                Annotation[] annotations, MediaType mediaType,
                MultivaluedMap<String, Object> httpHeaders, OutputStream entityStream)
                throws IOException, WebApplicationException {
            try {
                // NOTE: the channel should not close the stream. Jersey manages the stream.
                StreamingChannel out = new StreamingChannel(entityStream, null, false);
                out.sendMsg(xdrMessage);
            } catch (Exception e) {
                throw new ProcessingException("Error serializing XdrMessage", e);
            }
        }
    }

    public class HttpRpcClientProcessingException extends Exception {

        public HttpRpcClientProcessingException(Throwable t) {
            super(t);
        }
    }
}
