package devmgr.api.rpc;

/**************************************************************************
 *                                                                         *
 *  Copyright 2003-2011 NetApp, Inc. All Rights Reserved.                  *
 *                                                                         *
 *  This file is confidential and a trade secret of NetApp, Inc.  The      *
 *  receipt of or possession of this file does not convey any rights to    *
 *  reproduce or disclose its contents or to manufacture, use, or sell     *
 *  anything it may describe, in whole, or in part, without the specific   *
 *  written consent of NetApp, Inc.                                        *
 *                                                                         *
 **************************************************************************/

import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.Decoder;
import devmgr.api.runtime.XdrMessage;
import devmgr.api.symbol.SYMbolAPI;

import java.io.IOException;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.Date;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicLong;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A concrete implementation of {@link Client} that sends the RPC protocol. This class is not
 * thread safe. There are connect and call methods for sync or async approaches.
 */
public class NIORPCClient implements Client {
    private static final Logger logger = LoggerFactory.getLogger(NIORPCClient.class);
    private static AtomicLong symbolCallCounter = new AtomicLong();

    public static long getSymbolCallCount() {
        return symbolCallCount;
    }

    private static long symbolCallCount;

    private SelectManager mgr;
    private final XdrProtocolHelper protocolHelper;

    private int nextXid;

    private static final int DEFAULT_TIMEOUT = -1;

    final private String uuid; //random unique ID for this client
    private SocketChannel m_Channel;

    private AtomicLong lastTouched=new AtomicLong();

    private final Object callSync=new Object();

    //current call state
    private int xid;

    private byte[] header = new byte[4];
    private int bodyLen = -1;
    private int lengthCode=0;
    private int amountOfHeaderRead = 0;
    private int amountOfBodyRead = 0;

    private XdrMessage bodyBuffer;
    private byte[] currentBodySegment;
    private int timeout=10; //time in seconds
    private int userDefinedTimeout=DEFAULT_TIMEOUT;

    // Callback used to communicate success/failure/timeout; if this is null, there is no command in progress
    private NIOClientCallback currentCmdCallback;
    private int proc; //current call proc

    private Throwable lastError;
    private String controllerRef;
    private Throwable creationStack;
    private InetSocketAddress addr;
    /**
     * This class should not be created directly. Instead use {@link ClientTransportBuilder}.
     * Create the {@link SocketChannel} if the argument is null and
     * register the channel with the {@link SelectManager}. After this completes the channel is open
     * and registered with an nio selector, but it is not connected.
     *
     * @param mgr SelectKey channel manager for dispatching nio callbacks
     * @param channel the socket channel. This is usually null and a non-blocking channel will be
     * created.
     * @throws RPCError if there's an RPC error
     * @throws IOException if there's an IO issue
     */
    NIORPCClient(SelectManager mgr,SocketChannel channel) throws RPCError, IOException {
        creationStack=new Throwable("who created me");
        this.uuid=UUID.randomUUID().toString();
        protocolHelper=new XdrProtocolHelper(api);
        if (channel == null){
            m_Channel = SocketChannel.open();
            m_Channel.setOption(StandardSocketOptions.TCP_NODELAY,Boolean.TRUE);
            m_Channel.configureBlocking(false);
        }else{
            m_Channel=channel;
        }
        this.mgr=mgr;
        mgr.registerClient(this);


        // Set up the transaction ID sequencer using the current time
        // of day value from the system clock.
        nextXid = (int) ( (new Date()).getTime() / 1000 );
    }

    /**
     * Obtain the next valid transaction ID value for use with this RPC client object.
     *
     * @return The next transaction ID value to be used by this client.
     */
    private synchronized int getNextXid()
    {
        return nextXid++;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setTypeOverrides(Decoder.TypeOverrideTable typeOverrides) {
        this.protocolHelper.setTypeOverrides(typeOverrides);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server) throws Exception {
        connect(server, SYMbolAPI.PORT);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server, int port) throws Exception{
        logger.info("Creating a BLOCKING handler for the connection");
        final BlockingCallbackHandler callback=new BlockingCallbackHandler();
        connect(server, port, callback);
        // NOTE: This synchronizes on a local variable intentionally. The `connect` call is using
        // the blocking handler which will notifyAll() when it is complete.
        synchronized (callback){
            while (!callback.isDone()){
                callback.wait(500);
            }
        }
        if (callback.lastError != null){
            logger.info("Blocking connect failed with error", callback.lastError);
            throw callback.lastError;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void connect(InetAddress server, NIOClientCallback callback) {
        connect(server,SYMbolAPI.PORT,callback);
    }

    /**
     * {@inheritDoc} After this call completes, the channel will be initializing the connection. The
     * SelectManager thread uses java.nio.Selector, in {@link SelectManager#checkChannels(boolean)},
     * and notifies this class of connection state using the connection callbacks, such as
     * handleConnectSocketSignal.
     */
    @Override
    public void connect(InetAddress server,int port, NIOClientCallback callback) {
        lastTouched.set(System.currentTimeMillis());
        addr = new InetSocketAddress(server, port);
        // TODO: this should probably be in a sync block
        this.currentCmdCallback=callback;
        try{
            // The -1 procedureId indicates that we're initializing the socket
            proc=-1;
            m_Channel.connect(addr);
        }catch (Exception t){
            logger.info("Connect failed with error", t);
            lastError =t;
            if (currentCmdCallback != null){
                logger.debug("Issuing callback.onError");
                currentCmdCallback.onError(t);
            }
        }
    }


    @Override
    public int hashCode() {
        return uuid.hashCode();
    }



    @Override
    public boolean equals(Object obj) {
        boolean ret=false;
        if (obj instanceof NIORPCClient){
            NIORPCClient x=(NIORPCClient)obj;
            ret=uuid.equals(x.uuid);
        }
        return ret;
    }



    public SocketChannel getChannel() {
        return m_Channel;
    }



    long getLastTouched() {
        return lastTouched.get();
    }

    /**
     * Set a custom procedure timeout.
     *
     * @param timeoutInSeconds the custom procedure timeout (in seconds)
     */
    @Override
    public void setTimeout(int timeoutInSeconds)  {
        // NOTE: WARNING this does not set 'timeout', it sets 'userDefinedTimeout'
        logger.debug("Using custom timeout={} for proc={}", timeoutInSeconds, proc);
        userDefinedTimeout=timeoutInSeconds;
    }

    /**
     * Get the timeout in seconds. When a timeout occurs is described in {@link #handleTimeout()}.
     *
     * @return timeout of the command (seconds)
     */
    public int getTimeout(){
        return timeout;
    }

    /**
     * {@inheritDoc}
     * Unregister the channel from the select manager and close the networking channel.
     */
    @Override
    public void close() throws IOException {
        logger.debug("Removing registration and closing channel");
        mgr.deregisterClient(this);
        if (m_Channel != null){
            m_Channel.close();
            m_Channel = null;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public InetAddress getLocalAddress() {
        return m_Channel.socket().getLocalAddress();
    }

    /**
     * {@inheritDoc} This will return {@code null} if the channel is not
     * connected.
     */
    @Override
    public InetAddress getDestinationIP() {
        return m_Channel.socket().getInetAddress();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <T> T call(String proc, Object args) throws Exception{
        ApiDefinition.ProcTableEntry x = Client.api.getProcByName(proc);
        return call(x.procNumber, args);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <T> T call(int proc, Object args) throws Exception{
        final BlockingCallbackHandler callback=new BlockingCallbackHandler<T>();
        call(proc, args,callback);
        // NOTE: This synchronizes on a local variable intentionally. The `connect` call is using
        // the blocking handler which will notifyAll() when it is complete.
        synchronized (callback){
            while (!callback.isDone()){
                callback.wait(500);
            }
        }
        if (callback.lastError != null){
            throw callback.lastError;
        }
        return (T)callback.lastResults;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Object call(String proc, Object args, NIOClientCallback callback){
        ApiDefinition.ProcTableEntry x = Client.api.getProcByName(proc);
        return call(x.procNumber,args,callback);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int call(int proc, final Object args, NIOClientCallback callback){
        assert (lastError ==null);
        if (callback == null){
            throw new IllegalArgumentException("Callback cannot be null");
        }
        if (isActiveCall()){
            throw new IllegalStateException("Cannot run concurrent calls on a single client");
        }
        lastTouched.set(System.currentTimeMillis());
        symbolCallCount = symbolCallCounter.incrementAndGet();
        this.proc=proc;
        ApiDefinition.ProcTableEntry entry = api.getProcByNumber(proc);
        timeout = userDefinedTimeout == DEFAULT_TIMEOUT ? (entry.suggestedTimeout*2) : userDefinedTimeout;

        xid = getNextXid();
        // TODO: This should probably be in a synchronized block.
        this.currentCmdCallback=callback;
        try{
            logger.debug("sendMsg for {} proc={} timeout={} xid={}",
                    api.getProcByNumber(proc).procName, proc, timeout, xid);
            XdrMessage callMsg = protocolHelper.buildRequestMsg(proc, xid, args);
            resetCallBuffers();
            sendMsg(callMsg);
        }catch (Exception e){
            handleError(e);
        }

        return xid;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isConnected(){
        return m_Channel.isConnected();
    }


    private void resetCallBuffers() {
        // reset call buffers
        amountOfHeaderRead = 0;
        amountOfBodyRead = 0;
        bodyLen = -1;
        lengthCode=0;
        currentBodySegment=null;
        bodyBuffer = new XdrMessage();

    }

    /**
     * Read the xdr reply, which may not be the full message
     *
     * @throws IOException on error
     */
    private void processBytes() throws IOException {
        ByteBuffer readBuf = ByteBuffer.allocate(3096);
        int amtRead=m_Channel.read(readBuf);
        if (amtRead <0){
            handleError(new Exception("Unexpected Channel Close"));
        }
        byte[] buf=readBuf.array();
        while (amtRead > 0) {
            int amtToProcess = amtRead;
            // loop on the amount to read to process header and body if we dont
            // have all the header yet.
            while ( (amtToProcess > 0)&&  isActiveCall()) {
                if (amountOfHeaderRead != header.length) {
                    int n = Math.min(amtToProcess,
                            (header.length - amountOfHeaderRead));
                    int srcPos = amtRead - amtToProcess;
                    System.arraycopy(buf, srcPos, header, amountOfHeaderRead, n);
                    amountOfHeaderRead += n;
                    amtToProcess -= n;
                    if (amountOfHeaderRead == header.length){
                        // if we have read the header but not yet allocated the
                        // body, do it now
                        lengthCode = (((int) (header[0]) & 0xFF) << 24)
                                | (((int) (header[1]) & 0xFF) << 16)
                                | (((int) (header[2]) & 0xFF) << 8)
                                | (((int) (header[3]) & 0xFF) << 0);
                        bodyLen=lengthCode & 0x7FFFFFFF;
                        currentBodySegment=new byte[bodyLen];
                    }
                }
                // if did not eat all the bits in the header, keep going on the
                // body
                if (amtToProcess > 0) {// NOW READ THE BODY
                    // now process the body of the rest of the remaining bits as
                    // message body
                    int srcPos = amtRead - amtToProcess;
                    int lenToUse=Math.min(bodyLen-amountOfBodyRead, amtToProcess);
                    System.arraycopy(buf, srcPos, currentBodySegment,amountOfBodyRead, lenToUse);
                    amountOfBodyRead += lenToUse;
                    amtToProcess -=lenToUse;
                    if (amountOfBodyRead == bodyLen){
                        //reset the headers to read next chunk
                        amountOfHeaderRead=0;
                        bodyLen=-1;
                        amountOfBodyRead=0;
                        bodyBuffer.addBuffer(currentBodySegment);
                    }
                }

            }
            readBuf = ByteBuffer.allocate(3096);
            amtRead=m_Channel.read(readBuf);
            buf=readBuf.array();

        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setAuthGenerator(RPCAuthGenerator generator) {
        protocolHelper.setAuthGenerator(generator);
    }

    /**
     * {@inheritDoc}
     */
    public Throwable getLastError() {
        return lastError;
    }
    // ******************************* T A K E N F R O M T C P C H A N N E L C L A S S *************************************************************************

    /**
     * Send an XdrMessage to the remote host.
     *
     * @param msg the rpc message to execute on the server
     * @throws RPCError if there's an rpc error
     * @throws IOException if there's an io issue
     */
    public void sendMsg(final XdrMessage msg) throws RPCError, IOException {
        logger.debug("RPC: Sending message for proc={}", proc);
        final int totalSize = msg.getMessageSize();
        int amtDone = 0;
        int nextBuf = 0;
        //We will re-use this buffer
        final byte[] header = new byte[4];
        final ByteBuffer headerBuf = ByteBuffer.wrap(header);

        while (amtDone < totalSize) {
            final byte[] buf = msg.getBuffer(nextBuf++);
            final int amt = Math.min(buf.length, totalSize - amtDone);

            int lengthCode = amt;
            if ((amtDone + amt) >= totalSize) {
                lengthCode |= 0x80000000;
            }

            // set up the TCP record header
            header[0] = (byte) (lengthCode >>> 24);
            header[1] = (byte) (lengthCode >>> 16);
            header[2] = (byte) (lengthCode >>> 8);
            header[3] = (byte) (lengthCode);

            //Write the header first
            writeFully(headerBuf);
            //Write the real data out to the stream
            final ByteBuffer writeBuf = ByteBuffer.wrap(buf, 0, amt);
            writeFully(writeBuf);
            amtDone += amt;
        }

        // force transmission of any buffered data
        m_Channel.socket().getOutputStream().flush();
    }
    private void writeFully(ByteBuffer writeBuf) throws IOException {
        int n;
        writeBuf.rewind();
        n= m_Channel.write(writeBuf);
        int remaining = writeBuf.remaining();
        while (remaining>0){
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                // TODO Auto-generated catch block
                logger.error("An exception happened while writing on the channel", e);
            }
            n= m_Channel.write(writeBuf);
            remaining=writeBuf.remaining();

        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int getXid() {
        return xid;
    }

    /**
     * Mark the active call as being complete, indicating the channel is ready to service another
     * call. Note that this does not assume that the call succeeded or failed, only that it
     * completed. This causes {@link #isActiveCall()} to return false.
     */
    private void endCurrentCommand(){
        currentCmdCallback=null;
    }

    //*********** P A C K A G E    L E V E L   C A L L B A C K S from SELECTOR ************

    /**
     * Callback triggered when a timeout is detected on the channel. More specifically, every {@link
     * #call(int, Object, NIOClientCallback)} or {@link #connect(InetAddress, int,
     * NIOClientCallback)} will update {@code lastTouched} (an {@link AtomicLong}); the
     * SelectManager thread signals {@code onTimeout} when {@code (currentTime - lastTouched) >
     * timeout}.
     *
     * <p/>{@link SelectManager#checkChannels(boolean)} Selector thread calls this during its
     * channel processing.
     */
    synchronized void handleTimeout() {
        logger.info("Handle Timeout issued for proc={}", proc);
        lastError=new Exception("Timeout");
        if (currentCmdCallback != null){
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            NIOClientCallback tmp=currentCmdCallback;
            endCurrentCommand();
            tmp.onTimeout();
        }else{
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            endCurrentCommand();
        }


    }
    /**
     * Callback triggered when an error is detected on the channel.
     *
     * <p/>{@link SelectManager#checkChannels(boolean)} Selector thread calls this during its
     * channel processing.
     */
    synchronized void handleError(Exception t) {
        logger.info("Handle Error issued for proc={}", proc);
        logger.trace("Caller stack trace: ", creationStack);
        lastError=t;
        if (currentCmdCallback != null){
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            NIOClientCallback tmp=currentCmdCallback;
            endCurrentCommand();
            tmp.onError(t);
        }else{
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            endCurrentCommand();
        }

    }

    /**
     * Callback triggered when the channel has a pending connection
     * (see {@link java.nio.channels.SelectionKey#isConnectable()}).
     *
     * <p/>{@link SelectManager#checkChannels(boolean)} Selector thread calls this during its
     * channel processing.
     */
    synchronized void handleConnectSocketSignal() throws IOException {
        logger.debug("Handle Socket Signal issued for proc={}", proc);
        try {
            // Finish the connection setup. Note that handleConnectSocketSignal is called only when
            // the selection key says it has a connection pending, so there is no need to check the
            // channel for a pending connection (because we know it is pending, the key said so).
            //
            // Here's a quick summary of the important bits in the javadoc for finishConnect():
            // - It will throw an exception if there was an IO error
            // - It returns false if it didn't finish the connection. DO NOT call onSuccess since
            //      following read/write calls will fail. We'll try to finish the connection later.
            // - It returns true if it completed the connection or is already connected.
            if (!m_Channel.finishConnect()) {
                logger.debug("handleConnectSocketSignal called but the channel is not finished connecting");
                return;
            }

            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            NIOClientCallback tmp = currentCmdCallback;
            endCurrentCommand();
            tmp.onSuccess(null);
        } catch (Exception t) {
            logger.debug("Error on the channel while connecting to {}", addr.getHostString());
            lastError = t;

            NIOClientCallback tmp = currentCmdCallback;
            endCurrentCommand();
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            tmp.onError(t);
        }
    }

    /**
     * Callback triggered when the channel has data to read.
     *
     * <p/>{@link SelectManager#checkChannels(boolean)} Selector thread calls this during its
     * channel processing.
     *
     * @throws IOException if there's an I/O error
     */
    synchronized void handleReadSignal() throws IOException {
        logger.debug("Handling read signal for proc={}", proc);
        processBytes();
        synchronized (callSync) {
            if (currentCmdCallback == null){
                return;
            }
        }

        boolean doneCode=((lengthCode & 0x80000000) != 0) ;
        boolean allDataFetched=bodyLen == -1 ? true : amountOfBodyRead == bodyLen;
        boolean done=doneCode && allDataFetched;
        if (done ) {
            logger.debug("Completed reading RPC reply for proc={}", proc);
            NIOClientCallback cmdToUse=currentCmdCallback;
            try {
                logger.debug("Processing reply message and starting callback.onSuccess");
                Object result = protocolHelper.processReplyMsg(proc, bodyBuffer, xid);
                //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
                endCurrentCommand();
                cmdToUse.onSuccess(result);
            } catch (Exception e) {
                logger.debug("Failed processing reply msg. Starting callback.onError");
                lastError=e;
                //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
                endCurrentCommand();
                cmdToUse.onError(e);
            }

        }
    }

    /**
     * Check if the channel is currently being used, and a caller is waiting for its callback.
     *
     * @return true if an operation is in progress. False otherwise.
     */
    boolean isActiveCall() {
        boolean ret;
        synchronized (callSync){
            ret=currentCmdCallback != null;
        }
        return ret;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setControllerRef(String controllerRef) {
        this.controllerRef = controllerRef;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getControllerRef() {
        return controllerRef;
    }

    Throwable getCreationStack() {
        return creationStack;
    }
}
