package devmgr.api.rpc;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.XdrMessage;

/**
 * This class implements a TCP-based RPC client.  It connects to the
 * remote server, and then manages the call/reply stream for the
 * connection.  It is derived from RPCClientGeneric, and implements
 * abstract methods from the base class that implement the TCP-flavor of
 * of RPC client communication with a remote server.
 */
class RPCClientTCP extends RPCClientGeneric
{
    private static final int DEFAULT_TIMEOUT = 25;

    private TCPChannel m_Channel = null;

    /**
     * Constructor for TCP-based RPC clients.  Connects to the remote
     * server using a new stream socket, and prepares for ensuing
     * send/receive (i.e. call/return) sequences.
     * @param apiDef The API definition for the program.
     * @param server IP address of remote server.
     * @param port Explicit port number to connect to.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If an I/O error occurs.
     */
    RPCClientTCP(ApiDefinition apiDef, InetAddress server, int port)
    throws RPCError, IOException
    {
        super(apiDef);
        prepareConnection(server, port);
    }

    /**
     * Sets the timeout limit for the underlying connection.
     * @param seconds Number of seconds to wait before timing out.
     * @exception IOException If an I/O error occurs.
     */
    public void setTimeout(int seconds)
    throws IOException
    {
        m_Channel.setTimeout(seconds);
    }

    /**
     * Closes the underlying network connection.
     * @exception IOException If an I/O error occurs.
     */
    public void close()
    throws IOException
    {
        m_Channel.close();
        m_Channel = null;
    }

    /**
     * Returns the IP address of the client end of the RPC connection.
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress()
    {
        return m_Channel.getLocalAddress();
    }

    /**
     * Performs an RPC call/reply transaction for this connection.
     * Sends the specified call message, then waits for (and returns
     * to the caller with) the corresponding reply message.
     * @param callMsg The outbound call message.
     * @param xid The transaction ID for this exchange.
     * @return The IOV-based message containing reply data from the server.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If an I/O error occurs on the connection.
     */
    public XdrMessage transact(XdrMessage callMsg, int xid)
    throws RPCError, IOException
    {
        m_Channel.sendMsg(callMsg);
        return m_Channel.receiveMsg();
    }

    /**
     * Prepares a connection to the remote server.
     * @param server IP address of remote server.
     * @param port Port number to connect to.
     * @exception IOException If a connection or network error occurs.
     */
    private void prepareConnection(InetAddress server, int port)
    throws IOException
    {
        final Socket socket = new Socket(server, port);
        try {
            m_Channel = new TCPChannel(socket);
            setTimeout(DEFAULT_TIMEOUT);
        }
        catch(final IOException e){
            try {
                socket.close();
            }
            catch(final IOException ignored){
                //We'll want to swallow this so we can throw the top-level exception
            }
            throw e;
        }

    }

    /**
     * This method returns the address of the server to which this client
     * is connected.
     * @return Address of connected server
     */
    public InetAddress getDestinationIP()
    {
        return m_Channel.getServerAddress();
    }
    
}
