package devmgr.api.rpc;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.Date;
import java.util.concurrent.TimeUnit;

import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.XdrMessage;

/**
 * This class implements a UDP-based RPC client.  It sets up a socket
 * for communication with the remote server, and then manages the
 * call/reply stream for the connection.  It is derived from the
 * RPCClientGeneric base class, and implements abstract methods to
 * provide a UDP-flavor of RPC communication with a remote server.
 */
class RPCClientUDP extends RPCClientGeneric
{
    private static final int DEFAULT_TIMEOUT = 25;

    private long m_Timeout = DEFAULT_TIMEOUT * 1000;
    private InetAddress m_ServerAddress;
    private int m_ServerPort;
    private UDPChannel m_Channel = null;

    /**
     * Constructor for UDP-based RPC clients.  Creates the UDP channel,
     * and prepares for send/receive (i.e. call/return) sequences.  Uses
     * the specified port instead of interacting with a remote port
     * mapper to determine it.
     * @param apiDef The API definition for the program.
     * @param server IP address of remote server.
     * @param port Explicit port number to connect to.
     * @exception RPCError If an RPC error occurs.
     * @exception IOException If any I/O error occurs.
     */
    RPCClientUDP(ApiDefinition apiDef, InetAddress server, int port)
    throws RPCError, IOException
    {
        super(apiDef);

        // The caller has provided the remote server's IP address and port
        // number.  We blindly trust that these are valid, so we record them
        // and set up a UDP channel for ensuing network interactions.
        m_ServerAddress = server;
        m_ServerPort = port;

        m_Channel = new UDPChannel(new DatagramSocket());
    }

    /**
     * Sets the timeout limit for the underlying connection.
     * @param seconds Number of seconds to wait before timing out.
     * @exception IOException If an I/O error occurs.
     */
    public void setTimeout(int seconds)
    throws IOException
    {
        m_Timeout = TimeUnit.SECONDS.toMillis(seconds);

        // Set the underlying channel's timeout value, but only if the caller
        // is explicitly disabling the timeout mechanism.  For all non-zero
        // timeout values, we just set the member variable for use during I/O
        // operations.  We do it this way because the receive logic will always
        // loop waiting for a valid reply to arrive, so each loop iteration has
        // to adjust the timeout value of the underlying socket to account for
        // only the REMAINING time before a timeout should occur.  See the
        // transact() method for details on how this works.

        if (seconds == 0)
        {
            m_Channel.setTimeout(0);
        }
    }

    /**
     * Closes the underlying network connection.
     * @exception IOException If an I/O error occurs.
     */
    public void close()
    throws IOException
    {
        m_Channel.close();
        m_Channel = null;

        return;
    }

    /**
     * Returns the IP address of the client end of the RPC connection.
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress()
    {
        return m_Channel.getLocalAddress();
    }

    /**
     * Performs an RPC call/reply transaction for this connection.
     * Sends the specified call message, then waits for (and returns
     * to the caller with) the corresponding reply message.
     * @param callMsg The outbound call message.
     * @param xid The transaction ID for this exchange.
     * @return The IOV-based message containing reply data from the server.
     * @exception RPCError If any RPC error occurs.
     * @exception IOException If an I/O error occurs on the connection.
     */
    public XdrMessage transact(XdrMessage callMsg, int xid)
    throws RPCError, IOException
    {
        boolean done = false;
        UDPMessageBundle outBundle = null;
        UDPMessageBundle inBundle = null;
        XdrMessage replyMsg = null;
        Date startTime;
        Date nowTime;
        long timeWaited;

        // Turn the outbound message into a "bundle" that can be handled by
        // the underlying channel implementation, then pass it on for actual
        // network transmission.

        outBundle = new UDPMessageBundle(callMsg, m_ServerAddress, 
                                         m_ServerPort);
        m_Channel.sendMsg(outBundle);

        // Wait here for a valid reply message to arrive, subject to the
        // timeout value specified for the channel.  Note that the underlying
        // socket timeout value is reset each iteration through to account
        // for only the time REMAINING before a timeout should occur.
        startTime = new Date();
        while ( ! done )
        {
            if (m_Timeout > 0)
            {
                nowTime = new Date();
                timeWaited = nowTime.getTime() - startTime.getTime();

                if (timeWaited > m_Timeout)
                    throw new RPCError(RPCError.ErrorType.TIMEOUT);

                m_Channel.setTimeout((int)(m_Timeout - timeWaited) + 1);
            }

            // Here's where we actually wait for a reply message to arrive.
            inBundle = m_Channel.receiveMsg();
            replyMsg = inBundle.getMsg();

            // Check the XID before falling out of the loop; the high-level
            // code also checks this, so it may appear redundant to do it here.
            // However, if our check fails, we'll wait longer for more replies
            // to arrive, so there's a better chance of success.  Once we
            // return to the caller, there's no chance of getting a valid
            // reply at all.
            if (fetchXid(replyMsg) != xid)
                replyMsg = null;
            else
                done = true;
        }

        return replyMsg;
    }

    /**
     * Fetch the XID (transaction ID) from the specified message's
     * buffer.  Account for network byte ordering when extracting
     * the value from the message.
     * @param msg The message being processed.
     * @return The XID value to be used in the next call/reply transaction.
     */
    private int fetchXid(XdrMessage msg)
    {
        int xid;
        byte[] buf = msg.getBuffer(0);

        if (buf.length < 4)
        {
            xid = -1;
        }
        else
        {
            xid = ( ((int)(buf[0]) & 0xFF) << 24 ) |
                  ( ((int)(buf[1]) & 0xFF) << 16 ) |
                  ( ((int)(buf[2]) & 0xFF) <<  8 ) |
                  ( ((int)(buf[3]) & 0xFF) <<  0 ) ;
        }

        return xid;
    }

    /**
     * This method returns the address of the server to which this client
     * is connected.
     * @return Address of connected server
     */
    public InetAddress getDestinationIP()
    {
        return m_ServerAddress;
    }
}
