package devmgr.api.rpc;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;

import devmgr.api.runtime.XdrMessage;

/**
 * This class implements a generic RPC channel for use with streams.
 */
public class StreamingChannel {
    private OutputStream out;
    private InputStream in;
    private boolean closeStreams;

    /**
     * Constructor for the RPC channel using streams.
     *
     * @param out output stream for sending messages
     * @param in input stream for receiving messages
     */
    public StreamingChannel(OutputStream out, InputStream in) {
        this(out, in, true);
    }

    /**
     * Constructor for the RPC channel using streams.
     *
     * @param out output stream for sending messages
     * @param in input stream for receiving messages
     * @param closeStreams true for the IO stream to be closed by this class
     */
    StreamingChannel(OutputStream out, InputStream in, boolean closeStreams) {
        this.out = out;
        this.in = in;
        this.closeStreams = closeStreams;
    }

    /**
     * Closes the underlying network connection. This is a NOP and streams remain open if
     * {@code closeStreams} is set to false.
     *
     * @throws IOException If any I/O error occurs.
     */
    public void close() throws IOException {
        IOException exception = null;

        // If an external class is managing the streams, skip closing the stream.
        if (!this.closeStreams) {
            return;
        }

        try {
            in.close();
            in = null;
        } catch (IOException e) {
            exception = e;
        }

        try {
            out.close();
            out = null;
        } catch (IOException e) {
            exception = e;
        }

        if (exception != null) {
            throw exception;
        }
    }

    /**
     * Sends an outbound RPC message.
     *
     * @param msg The message to be sent.
     * @throws RPCError If an RPC error occurs.
     * @throws IOException If any I/O error occurs.
     */
    public void sendMsg(XdrMessage msg) throws RPCError, IOException {
        int totalSize = msg.getMessageSize();
        int amtDone = 0;
        int nextBuf = 0;
        byte[] header = new byte[4];

        while (amtDone < totalSize) {
            byte[] buf = msg.getBuffer(nextBuf++);
            int amt = Math.min(buf.length, totalSize - amtDone);

            int lengthCode = amt;
            if ((amtDone + amt) >= totalSize) {
                lengthCode |= 0x80000000;
            }

            // Set up the TCP record header for this chunk.
            header[0] = (byte) (lengthCode >>> 24);
            header[1] = (byte) (lengthCode >>> 16);
            header[2] = (byte) (lengthCode >>> 8);
            header[3] = (byte) (lengthCode >>> 0);

            // Write the header and data to the output stream.
            out.write(header);
            out.write(buf, 0, amt);

            amtDone += amt;
        }

        // Force transmission of any buffered data.
        out.flush();
    }

    /**
     * Receives an inbound RPC message and returns it to the caller.
     *
     * @return The received message
     * @throws RPCError If an RPC error occurs.
     * @throws IOException If any I/O error occurs.
     */
    public XdrMessage receiveMsg() throws RPCError, IOException {
        XdrMessage msg = new XdrMessage();
        byte[] header = new byte[4];
        boolean done = false;
        int amt;

        while (!done) {
            readFully(header, 0, header.length);
            int lengthCode = (((int)(header[0]) & 0xFF) << 24) |
                    (((int)(header[1]) & 0xFF) << 16) |
                    (((int)(header[2]) & 0xFF) <<  8) |
                    (((int)(header[3]) & 0xFF) <<  0) ;

            if ((amt = (lengthCode & 0x7FFFFFFF)) != 0) {
                byte[] buf = new byte[amt];
                readFully(buf, 0, amt);
                msg.addBuffer(buf);
            }

            done = ((lengthCode & 0x80000000) != 0);
        }

        return msg;
    }

    /**
     * Receives the specified amount of data from the connection,
     * possibly iterating and accumulating partial results.
     *
     * @param buf Buffer into which data is placed.
     * @param off Buffer offset at which to start filling.
     * @param amt Amount of data to receive.
     * @throws RPCError If an RPC error occurs.
     * @throws IOException If any I/O error occurs.
     */
    private void readFully(byte[] buf, int off, int amt) throws RPCError, IOException {
        int done = 0;
        int n;

        while (done < amt) {
            //Use a try/catch block so we can correctly report RPC TIMEOUTs.
            try {
                n = in.read(buf, off + done, amt - done);
            } catch (InterruptedIOException ex) {
                throw new RPCError(RPCError.ErrorType.TIMEOUT);
            }

            if (n <= 0) {
                throw new RPCError(RPCError.ErrorType.CONNECTION_LOST);
            }

            done += n;
        }
    }
}
