package devmgr.api.rpc;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

/**
 * This class implements a TCP channel for carrying RPC messages.
 * Its methods are responsible for adding/removing RPC 4-byte
 * message length codes, etc., so as to provide the user with
 * a more abstract channel that is used to send/receive basic
 * messages without worrying about the TCP-oriented details.
 */
public class TCPChannel extends StreamingChannel
{
    private Socket m_Sock = null;

    /**
     * Constructor for the RPC channel class for TCP connections.
     * Just prepares local resources for future use.
     * @param sock Socket to be used for the channel.
     * @exception IOException If any I/O error occurs.
     */
    public TCPChannel(Socket sock)
            throws IOException
    {
        super(sock.getOutputStream(), sock.getInputStream());

        m_Sock = sock;
        m_Sock.setTcpNoDelay(true);
    }

    /**
     * Sets the timeout limit for the underlying connection.
     * @param seconds Number of seconds to wait before timing out.
     * @exception IOException If any I/O error occurs.
     */
    public void setTimeout(int seconds)
            throws IOException
    {
        m_Sock.setSoTimeout(seconds * 1000);
        return;
    }

    /**
     * Closes the underlying network connection.
     * @exception IOException If any I/O error occurs.
     */
    public void close()
            throws IOException
    {
        IOException exception = null;
        try {
            super.close();
        } catch (IOException e) {
            exception = e;
        }

        try{
            m_Sock.close();
            m_Sock = null;
        } catch (IOException e) {
            exception = e;
        }

        if (exception != null) {
            throw exception;
        }

        return;
    }

    /**
     * Returns the IP address of the client end of the RPC connection.
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress()
    {
        return m_Sock.getLocalAddress();
    }

    /**
     * Get the InetAddress of the remote (server) system to which this
     * channel is connected.
     * @return Remote server address.
     */
    public InetAddress getServerAddress()
    {
        return m_Sock.getInetAddress();
    }
}
