package devmgr.api.security;

import java.security.GeneralSecurityException;
import java.security.SecureRandom;

import devmgr.api.runtime.ApiUtil;
import devmgr.api.security.PbeWithDigestAndCipher.PbeType;
import devmgr.api.symbol.EncryptedPassword;
import devmgr.api.symbol.EncryptionType;
import devmgr.api.symbol.PKCS5EncryptionData;
import devmgr.api.symbol.SAAttributeID;
import devmgr.api.symbol.SAData;
import devmgr.api.symbolutils.ByteUtils;

/**
 * Factory for generating EncryptedPasswords.
 * <P>
 * REFERENCES: <BR>
 * http://ictwiki.eng.netapp.com/index.php/SYMbol_security_enhancement
 */
public class EncryptedPasswordFactory {
    private static final int PBE_DES_ITERATIONS = 1000;
    private static final int PBE_AES_ITERATIONS = 10000;

    /**
     * Creates a new EncryptedPasswordFactory.
     * <P>
     * Reduced scope to prevent external instantiation.
     */
    private EncryptedPasswordFactory() {}

    /**
     * Builds an EncryptedPassword suitable for sending to the storage array.
     * <P>
     * Different password types have different encoding requirements.  Storage Array passwords
     * are encoded in a subset of UTF16-BE (UCS-2).  Most other passwords are encoded in US-ASCII.
     *
     * @param saData current storage array data, used to determine supported encryption type
     * @param currentArrayPassword the current encoded storage array password
     * @param passwordToEncrypt the encoded password to encrypt
     * @return encrypted password object suitable to send to the array
     * @throws GeneralSecurityException if an error occurs generating the encrypted data
     */
    public static EncryptedPassword buildEncryptedPassword(SAData saData,
            byte[] currentArrayPassword, byte[] passwordToEncrypt) throws GeneralSecurityException {
        // ArrayAttributes is an attribute followed by a value so we skip 2 to ensure we aren't
        // checking any values.
        for (int i = 0; i < saData.getArrayAttributes().length; i += 2) {
            if (saData.getArrayAttributes()[i] == SAAttributeID.symbolAuthHmac256Support.value) {
                return buildEncryptedPasswordSHA256AndAES(currentArrayPassword, passwordToEncrypt);
            }
        }

        // Didn't determine any other type, fall back to legacy mode
        return buildEncryptedPasswordSHA1AndDES(currentArrayPassword, passwordToEncrypt);
    }

    /**
     * Builds an encrypted password using the legacy PBE with SHA-1 and DES.
     *
     * @param currentArrayPassword the current encoded storage array password
     * @param passwordBytes encoded password to encrypt
     * @return encrypted password object suitable to send to the array
     * @throws GeneralSecurityException if an error occurs generating the encrypted data
     */
    private static EncryptedPassword buildEncryptedPasswordSHA1AndDES(byte[] currentArrayPassword,
            byte[] passwordBytes) throws GeneralSecurityException {
        PbeWithDigestAndCipher pbe = new PbeWithDigestAndCipher(PbeType.PKCS5_SHA1_DES);

        SecureRandom secureRandom = new SecureRandom();
        byte[] salt = new byte[pbe.getSaltSize()];
        secureRandom.nextBytes(salt);

        pbe.init(true, currentArrayPassword, salt, PBE_DES_ITERATIONS);

        EncryptedPassword encryptedPassword = new EncryptedPassword();
        encryptedPassword.setType(EncryptionType.pkcs5Sha1Des);
        encryptedPassword.setPkcs5Data(new PKCS5EncryptionData());
        encryptedPassword.getPkcs5Data().setIterations(PBE_DES_ITERATIONS);
        encryptedPassword.getPkcs5Data().setSalt(ApiUtil.opaqueToString(salt));
        encryptedPassword.getPkcs5Data().setCipherText(ApiUtil.opaqueToString(pbe.crypt(passwordBytes)));

        return encryptedPassword;
    }

    /**
     * Builds an encrypted password using the PBE with SHA-256 and AES.
     *
     * @param currentArrayPassword the current encoded storage array password
     * @param passwordBytes encoded password to encrypt
     * @return encrypted password object suitable to send to the array
     * @throws GeneralSecurityException if an error occurs generating the encrypted data
     */
    private static EncryptedPassword buildEncryptedPasswordSHA256AndAES(byte[] currentArrayPassword,
            byte[] passwordBytes) throws GeneralSecurityException {
        PbeWithDigestAndCipher pbe = new PbeWithDigestAndCipher(PbeType.PKCS5_SHA256_AES);

        SecureRandom secureRandom = new SecureRandom();
        byte[] salt = new byte[pbe.getSaltSize()];
        secureRandom.nextBytes(salt);

        pbe.init(true, currentArrayPassword, salt, PBE_AES_ITERATIONS);

        EncryptedPassword encryptedPassword = new EncryptedPassword();
        encryptedPassword.setType(EncryptionType.pkcs5Sha2Aes);
        encryptedPassword.setPkcs5AesData(new PKCS5EncryptionData());
        encryptedPassword.getPkcs5AesData().setIterations(PBE_AES_ITERATIONS);
        encryptedPassword.getPkcs5AesData().setSalt(ApiUtil.opaqueToString(salt));
        encryptedPassword.getPkcs5AesData().setCipherText(ApiUtil.opaqueToString(pbe.crypt(passwordBytes)));

        return encryptedPassword;
    }

    /**
     * Reverses the PBE to determine the encrypted password.
     *
     * @param encryptedPassword SYMbol data structure holding the encrypted password
     * @param encryptionKey key used to encrypt the password, usually the current array password
     * @return decrypted password
     * @throws GeneralSecurityException if an error occurs decrypting the encrypted data
     */
    public static byte[] decryptPassword(EncryptedPassword encryptedPassword, byte[] encryptionKey)
            throws GeneralSecurityException {
        PbeWithDigestAndCipher pbe;
        byte[] salt;
        int iterationCount;
        byte[] cypherText;

        switch (encryptedPassword.getType()) {
            case pkcs5Sha1Des:
                pbe = new PbeWithDigestAndCipher(PbeType.PKCS5_SHA1_DES);
                salt = ByteUtils.hexStringToByteArray(encryptedPassword.getPkcs5Data().getSalt());
                iterationCount = encryptedPassword.getPkcs5Data().getIterations();
                cypherText = ByteUtils.hexStringToByteArray(encryptedPassword.getPkcs5Data().getCipherText());
                break;
            case pkcs5Sha2Aes:
                pbe = new PbeWithDigestAndCipher(PbeType.PKCS5_SHA256_AES);
                salt = ByteUtils.hexStringToByteArray(encryptedPassword.getPkcs5AesData().getSalt());
                iterationCount = encryptedPassword.getPkcs5AesData().getIterations();
                cypherText = ByteUtils.hexStringToByteArray(encryptedPassword.getPkcs5AesData().getCipherText());
                break;
            default:
                throw new GeneralSecurityException("Unable to decrypt encryption type " + encryptedPassword.getType());
        }

        pbe.init(false, encryptionKey, salt, iterationCount);
        byte[] decryptedPassword = pbe.crypt(cypherText);
        return decryptedPassword;
    }
}
