package devmgr.api.security;

import devmgr.api.symbol.SAIdentifier;
import devmgr.api.symbolutils.ByteUtils;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;

/**
 * Security related utility methods which are specific to NetApp, Inc. storage array.
 * <P>
 * REFERENCE: <BR>
 * <a href="https://agile.lsi.com/
 * Agile/PCMServlet?module=ItemHandler&requestUrl=module%3DItemHandler%26opcode%3D
 * displayObject%26classid%3D9000%26objid%3D158673076%26tabid%3D11%26">FAM - 35598-00 revision
 *  A.2 </a><br>
 * FFD - 34200-00
 * ER_ViewOnlyPasswordManagementA4_CB.pdf : 53791-00
 * DA_ViewOnlyPasswordManagementA4_CB.pdf : 53887-00
 * UX-Dual Password UX design document : <a href=http://sharepoint.corp.netapp.com/sites/eng/
 * storsys/CFW/main/Boxcars/OCR_Galena/Lists/TeamDiscussion UX Design Specification </a>
 */
public final class SymbolSecurity {
    /**
     * Encrypt the pass phrase. The key encryption key and the pass phrase are hashed using SHA-256.
     * The hashed pass phrase is wrapped using AES wrap with the hashed key.
     *
     * @param passPhrase The pass phrase to encrypt.
     * @param storageArrayPassword The user password of the storage array.
     * @param saIdentifier The storage array ID.
     * @return The wrapped pass phrase.
     * @throws GeneralSecurityException if the cipher cannot be init'ed or if an error occurs
     *         wrapping the passphrase.
     */
    public static byte[] encryptFDEPassPhrase(String passPhrase, String storageArrayPassword, SAIdentifier saIdentifier)
            throws GeneralSecurityException {
        MessageDigest sha256 = MessageDigest.getInstance("SHA-256");
        byte[] hashedPassPhrase = sha256.digest(passPhrase.getBytes(StandardCharsets.UTF_8));
        SecretKey keyEncryptionKey = generateKeyEncryptionKey(storageArrayPassword, saIdentifier);
        return encryptFDEPassPhrase(hashedPassPhrase, keyEncryptionKey);
    }

    /**
     * Validate the secure message with the given digest and salt.
     * The message digest is prepared by prefixing the salt to the beginning of the password.
     *
     * @param secureMessage secure message which needs to be validated.
     * @param salt Randomly generated salt of 8 bytes length
     * @param digest Message digest
     * @return true for a valid secure message or else false.
     * @throws java.security.NoSuchAlgorithmException if SHA digest not available
     */
    public static boolean validateSecureMessage(String secureMessage, byte[] salt, byte[] digest)
            throws NoSuchAlgorithmException {
        MessageDigest sha = MessageDigest.getInstance("SHA");
        sha.update(salt, 0, 8);
        byte[] bytes;
        try {
            bytes = secureMessage.getBytes("UTF-16BE");
        } catch (UnsupportedEncodingException e) {
            bytes = secureMessage.getBytes();
        }
        sha.update(bytes);
        byte[] messageDigest = sha.digest();
        return Arrays.equals(messageDigest, digest);
    }

    /**
     * Decrypts and then encrypts the FDE passphrase.
     *
     * @param encryptedPassPhrase encrypted pass phrase to re-encrypt
     * @param originalArrayPassword storage array password to use for decryption
     * @param newArrayPassword storage array password to use for encryption
     * @param saIdentifier used as part of the key encryption key if either storage array password is empty
     * @return re-encrypted FDE passphrase
     * @throws GeneralSecurityException if an error occurs during the decryption or encryption
     */
    public static byte[] reencryptFDEPassPhrase(byte[] encryptedPassPhrase, String originalArrayPassword, String newArrayPassword,
            SAIdentifier saIdentifier) throws GeneralSecurityException {
        SecretKey keyEncryptionKey = generateKeyEncryptionKey(originalArrayPassword, saIdentifier);
        Cipher cipher = Cipher.getInstance("AESWrap");
        cipher.init(Cipher.UNWRAP_MODE, keyEncryptionKey);
        Key unwrappedKey = cipher.unwrap(encryptedPassPhrase, "AES", Cipher.SECRET_KEY);

        keyEncryptionKey = generateKeyEncryptionKey(newArrayPassword, saIdentifier);
        return encryptFDEPassPhrase(unwrappedKey.getEncoded(), keyEncryptionKey);
    }

    /**
     * Encrypts a hashed FDE pass phrase.
     *
     * @param hashedPassPhrase hashed pass phrase
     * @param keyEncryptionKey key used for the encryption of the pass phrase
     * @return encrypted form of the pass phrase
     * @throws GeneralSecurityException if an error occurs encrypting the data
     */
    private static byte[] encryptFDEPassPhrase(byte[] hashedPassPhrase, SecretKey keyEncryptionKey)
            throws GeneralSecurityException {
        // Create the secret key out of the hashed pass phrase that Java needs for wrapping.
        SecretKey passPhraseSecretKey = new SecretKeySpec(hashedPassPhrase, "AES");

        // Get an AES key wrap cipher.
        Cipher cipher = Cipher.getInstance("AESWrap");

        cipher.init(Cipher.WRAP_MODE, keyEncryptionKey);
        return cipher.wrap(passPhraseSecretKey);
    }

    /**
     * Builds the key encryption key used for wrapping the passphrase.
     *
     * @param arrayPassword storage array password, if non-zero length this will be used for the key encryption key
     * @param saIdentifier if storage array password is empty, the WWN found in this will be used for the key encryption key
     * @return key encryption key to be used for FDE passphrase encryption / decryption
     * @throws GeneralSecurityException if an error occurs generating the key encryption key
     */
    private static SecretKey generateKeyEncryptionKey(String arrayPassword, SAIdentifier saIdentifier)
            throws GeneralSecurityException {
        byte[] keyEncryptionKey = arrayPassword == null || arrayPassword.isEmpty() ?
                ByteUtils.hexStringToByteArray(saIdentifier.getWorldWideName()) :
                arrayPassword.getBytes(StandardCharsets.UTF_16BE);

        //Hash the key encryption key with SHA-256.
        MessageDigest sha256 = MessageDigest.getInstance("SHA-256");
        byte[] hashedKeyEncryptionKey = sha256.digest(keyEncryptionKey);
        byte[] hashedKeyEncryptionKey128 = new byte[16];

        //Truncate to the first 128 bits.
        System.arraycopy(hashedKeyEncryptionKey, 0, hashedKeyEncryptionKey128, 0,
                hashedKeyEncryptionKey128.length);

        // Create the secret key out of the hashed password or SA ID that Java needs for wrapping.
        return new SecretKeySpec(hashedKeyEncryptionKey128, "AES");
    }
}
