package devmgr.api.workers;

import devmgr.api.model.*;
import devmgr.api.rpc.SymbolClient;
import devmgr.api.rpc.NIOClientCallback;
import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.Decoder;
import devmgr.api.symbol.*;
import devmgr.api.symbolutils.ByteUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;

/**
 * Created with IntelliJ IDEA.
 * User: masonb
 * Date: 11/30/13
 * Time: 4:03 PM
 * To change this template use File | Settings | File Templates.
 */
public class ObjectGraphFetcher implements AbstractObjectGraphFetcher {
    //constants to help with repository volumes
    private static final String REPOS_PREFIX = "repos_";
    private static final int REPOS_DIGITS_COUNT = 4;
    private static final String REPOS_REGEX = REPOS_PREFIX + "[0-9]{" + REPOS_DIGITS_COUNT + "}";
    static SYMbolAPI api = new SYMbolAPI();
    private String systemId; //never used in this version, but kept if needed
    Decoder.TypeOverrideTable typeOverrides =
            new Decoder.TypeOverrideTable();

    final Logger debugLogger;

    public ObjectGraphFetcher() {
        this(LoggerFactory.getLogger(ObjectGraphFetcher.class));
    }

    public ObjectGraphFetcher(@Deprecated final java.util.logging.Logger debugLogger) {
        this(LoggerFactory.getLogger(ObjectGraphFetcher.class));
    }

    public ObjectGraphFetcher(final Logger debugLogger) {
        this.debugLogger = debugLogger;
        typeOverrides.put(Volume.class, VolumeEx.class);
        typeOverrides.put(VolumeGroup.class, VolumeGroupEx.class);
        typeOverrides.put(FibreInterface.class, devmgr.api.model.FibreInterfacePort.class);
        typeOverrides.put(IscsiInterface.class,IScsiInterfacePort.class);
        typeOverrides.put(SasInterface.class, SasInterfacePort.class);
        typeOverrides.put(IbInterface.class, IBInterfacePort.class);
        typeOverrides.put(Host.class, devmgr.api.model.HostEx.class);
        typeOverrides.put(HostSpecificValues.class, HostTypeValues.class);
        typeOverrides.put(Cluster.class, HostGroup.class);
        typeOverrides.put(ThinVolume.class, ThinVolumeEx.class);
        typeOverrides.put(PITView.class, PitViewEx.class);
        typeOverrides.put(AccessVolume.class, AccessVolumeEx.class);
        typeOverrides.put(SnapshotVolume.class, LegacySnapshotEx.class);
        typeOverrides.put(Drive.class, DriveEx.class);
        typeOverrides.put(Tray.class, TrayEx.class);
        typeOverrides.put(VolumeCopy.class, VolumeCopyPair.class);
        typeOverrides.put(PIT.class, Snapshot.class);
        typeOverrides.put(PITGroup.class, SnapshotGroup.class);
        typeOverrides.put(Alarm.class, AlarmEx.class);
        typeOverrides.put(EthernetInterface.class, ManagementInterface.class);
        typeOverrides.put(AsyncMirrorGroup.class, Amg.class);
        typeOverrides.put(AsyncMirrorGroupMember.class, AmgMember.class);
        typeOverrides.put(AsyncMirrorGroupIncompleteMember.class, AmgIncompleteMember.class);
        typeOverrides.put(FlashCache.class, FlashCacheEx.class);
        typeOverrides.put(ConcatVolume.class, ConcatRepositoryVolume.class);
        typeOverrides.put(Battery.class, BatteryEx.class);
    }

    @Override
    public void setSystemId(String id) {
        this.systemId=id;
    }

    private Set<ProductCapability> capabilities = Collections.emptySet();

    @Override
    public void setProductCapabilities(Set<ProductCapability> capabilities) {
        this.capabilities = capabilities;
    }

    /**
     * Fetches the object graph, blocking if callback is null, otherwise it runs async
     * @param client
     * @param callback
     */
    @Override
    public ObjectBundle doObjectGraphFetch(final ExecutorService pool, final SymbolClient client, final NIOClientCallback<ObjectBundle> callback) throws
            Exception {

        NIOClientCallback realCallback=new NIOClientCallback<ObjectBundle>(){
            @Override
            public void onTimeout() {
                callback.onTimeout();
            }

            @Override
            public void onError(Exception t) {

                debugLogger.info("Error fetching Object Graph: ", t);
                callback.onError(t);
            }

            @Override
            public void onSuccess(final ObjectBundle results) {
                pool.submit(() -> {
                    try {
                        augmentObjects(results);
                        callback.onSuccess(results);
                    }
                    catch(Exception e){
                        callback.onError(e);
                    }
                });

            }
        };
        client.setTypeOverrides(typeOverrides);
        ApiDefinition.ProcTableEntry proc = api.getProcByName("getObjectGraph");
        NIOClientCallback x=callback != null ? realCallback : null;
        ObjectBundle ret;
        if (x == null){
            ret= (ObjectBundle)client.call(proc.procNumber, null);
            augmentObjects(ret);
        }else{
            client.call(proc.procNumber,null,realCallback);
            ret=null;
        }
        return ret;
    }

    private void augmentManagementInterfaces(ObjectBundle obj){
        try {
            for (Controller c : obj.getController()) {
                for (NetInterfaceTypeData iface : c.getNetInterfaces()) {
                    ManagementInterface mgmt = (ManagementInterface) iface.getEthernet();
                    ControllerNetworkSettings networkSettings = c.getNetworkSettings();
                    if (networkSettings != null) {
                        mgmt.setIpv4GatewayAddress(networkSettings.getIpv4DefaultRouterAddress());
                        mgmt.setDnsProperties(networkSettings.getDnsProperties());
                        mgmt.setNtpProperties(networkSettings.getNtpProperties());
                    }
                    mgmt.setControllerRef(c.getControllerRef());
                    mgmt.setControllerSlot(c.getPhysicalLocation().getSlot());
                }
            }
        }catch(Throwable t){
            debugLogger.info("Exception caught augmenting management interfaces!", t);
        }
    }

    private void augmentObjects(ObjectBundle obj) {
        augmentSa(obj);
        augmentSaData(obj);
        augmentVolume(obj);
        augmentVolumeGroups(obj);
        augmentConcatVolumes(obj);
        augmentInterface(obj);
        augmentHosts(obj);
        augmentHostGroup(obj);
        augmentThinVolumes(obj);
        augmentPitViews(obj);
        augmentLegacySnapshots(obj);
        augmentPitGroups(obj);
        augmentPits(obj);
        augmentHostTypeValues(obj);
        augmentManagementInterfaces(obj);
        augmentArvm(obj);
        augmentTrays(obj);
        augmentVolumeCopies(obj);
        augmentFlashCache(obj);
        augmentDrive(obj);
        augmentBattery(obj);
    }

    private void augmentSa(ObjectBundle obj) {
        StorageArray storageArray = obj.getSa();
        ChannelPort [] ports = storageArray.getChannelPortLocations();
        if(ports == null){
            ports = new ChannelPort[0];
            storageArray.setChannelPortLocations(ports);
        }
    }

    private Map<String, UserVolume> getUserVolumeMap(ObjectBundle bundle){
        VolumeEx [] volumes = (VolumeEx[]) bundle.getVolume();
        ThinVolumeEx [] thinVolumes = (ThinVolumeEx[]) bundle.getHighLevelVolBundle().getThinVolume();
        Map<String, UserVolume> ret = newHashMapWithExpectedSize(volumes.length + thinVolumes.length);
        for(ThinVolumeEx v : thinVolumes){
            ret.put(v.getId(), v);
        }
        for(VolumeEx v : volumes){
            ret.put(v.getId(), v);
        }
        return ret;
    }

    private Map<String, LegacySnapshotEx> getLegacySnapshotMap(ObjectBundle bundle){
        LegacySnapshotEx [] legacySnapshots = (LegacySnapshotEx[]) bundle.getHighLevelVolBundle().getSnapshotVol();
        Map<String, LegacySnapshotEx> ret = newHashMapWithExpectedSize(legacySnapshots.length);
        for(LegacySnapshotEx snap : legacySnapshots){
            ret.put(snap.getId(), snap);
        }
        return ret;
    }

    private Map<String, PitViewEx> getSnapshotViewMap(ObjectBundle bundle){
        PitViewEx [] views = (PitViewEx[]) bundle.getHighLevelVolBundle().getPitView();
        Map<String, PitViewEx> ret = newHashMapWithExpectedSize(views.length);
        for(PitViewEx snap : views){
            ret.put(snap.getId(), snap);
        }
        return ret;
    }


    private void augmentVolumeCopies(ObjectBundle bundle) {
        VolumeCopyPair [] pairs = (VolumeCopyPair[]) bundle.getHighLevelVolBundle().getVolumeCopy();
        Map<String, PitViewEx> viewMap = getSnapshotViewMap(bundle);
        Map<String, LegacySnapshotEx> legacySnapshotMap = getLegacySnapshotMap(bundle);

        //Volume copies are not supported
        if(pairs == null){
            bundle.getHighLevelVolBundle().setVolumeCopy(new VolumeCopyPair[0]);
        }
        //VolumeCopies are supported on the device
        else{
            for(VolumeCopyPair pair : pairs){
                final String sourceId = pair.getSourceVolume();
                PitViewEx view = viewMap.get(sourceId);
                LegacySnapshotEx snap = legacySnapshotMap.get(sourceId);
                if(view != null){
                    pair.setBaseSourceVolumeId(view.getBaseVol());
                }
                else if(snap != null){
                    pair.setBaseSourceVolumeId(snap.getBaseVolume());
                }
                //This is a normal volumeCopy, so the base is also the source
                else{
                    pair.setBaseSourceVolumeId(sourceId);
                }
            }
        }
    }

    //The default load factor for HashMaps and Sets
    private static final float LOAD_FACTOR = 0.75f;

    /**
     * Define a new HashMap with the capacity optimized to expect up to expectedSize elements.
     * @param expectedSize The expected size of the HashMap
     * @param <K> The Key type.
     * @param <V>
     * @return
     */
    private <K,V> Map<K, V> newHashMapWithExpectedSize(final int expectedSize){
        return new HashMap<>((int) Math.ceil(expectedSize / LOAD_FACTOR), LOAD_FACTOR);
    }

    /**
     * Define a new HashSet with the capacity optimized to expect up to expectedSize elements.
     * @param expectedSize The expected size of the HashSet
     * @param <K> The Key type.
     * @return
     */
    private <K> Set<K> newHashSetWithExpectedSize(final int expectedSize){
        return new HashSet<>((int) Math.ceil(expectedSize / LOAD_FACTOR), LOAD_FACTOR);
    }

    private void augmentConcatVolumes(ObjectBundle graph) {
        ConcatRepositoryVolume[] concatVolumes = (ConcatRepositoryVolume[]) graph.getHighLevelVolBundle().getConcatVolume();
        ConcatVolumeMember[] concatVolumeMembers = graph.getHighLevelVolBundle().getConcatVolMember();

        if(concatVolumes == null){
            concatVolumes = new ConcatRepositoryVolume[0];
            graph.getHighLevelVolBundle().setConcatVolume(concatVolumes);
        }
        if(concatVolumeMembers == null){
            concatVolumeMembers = new ConcatVolumeMember[0];
            graph.getHighLevelVolBundle().setConcatVolMember(concatVolumeMembers);
        }

        if(concatVolumes.length > 0) {
            ThinVolumeEx[] thinVolumes = (ThinVolumeEx[]) graph.getHighLevelVolBundle().getThinVolume(); 
            SnapshotGroup[] snapshotGroups = (SnapshotGroup[]) graph.getHighLevelVolBundle().getPitGroup();
            PitViewEx[] pitViews = (PitViewEx[]) graph.getHighLevelVolBundle().getPitView();
            AmgMember[] amgMembers = (AmgMember[]) graph.getHighLevelVolBundle().getAsyncMirrorGroupMember();

            Map<String,ConcatRepositoryVolume> concatMap = newHashMapWithExpectedSize(concatVolumes.length);
            for(ConcatRepositoryVolume repo : concatVolumes){
                concatMap.put(repo.getConcatVolRef(), repo);
            }

            for(ConcatVolumeMember member : concatVolumeMembers){
                final String concatVolumeRef = member.getConcatVolRef();
                final ConcatRepositoryVolume concatVolume = concatMap.get(concatVolumeRef);
                if(concatVolume != null){
                    List<String> keys = concatVolume.getMemberRefs();
                    if(keys == null){
                        keys = new ArrayList<>();
                        concatVolume.setMemberRefs(keys);
                    }
                    keys.add(member.getStorageVolumeRef());
                }
            }
            
            for(ThinVolumeEx thinVolume : thinVolumes) {
               final String concatVolumeRef = thinVolume.getRepositoryRef();
               final ConcatRepositoryVolume concatVolume = concatMap.get(concatVolumeRef);
               if(concatVolume != null) {
                  concatVolume.setBaseObjectId(thinVolume.getId());
                  concatVolume.setBaseObjectType(ObjectType.thinVolume);
               }
            }
            
            for(SnapshotGroup sg : snapshotGroups) {
               final String concatVolumeRef = sg.getRepositoryVolume();
               final ConcatRepositoryVolume concatVolume = concatMap.get(concatVolumeRef);
               if(concatVolume != null) {
                  concatVolume.setBaseObjectId(sg.getId());
                  concatVolume.setBaseObjectType(ObjectType.snapshotGroup);
               }
            }
            
            for(AmgMember amgMember : amgMembers) {
               final String concatVolumeRef = amgMember.getRepositoryVolume();
               final ConcatRepositoryVolume concatVolume = concatMap.get(concatVolumeRef);
               if(concatVolume != null) {
                  concatVolume.setBaseObjectId(amgMember.getId());
                  concatVolume.setBaseObjectType(ObjectType.asyncMirrorGroupMember);
               }
            }
            
            for(PitViewEx view : pitViews) {
               final String concatVolumeRef = view.getRepositoryVolume();
               final ConcatRepositoryVolume concatVolume = concatMap.get(concatVolumeRef);
               if(concatVolume != null) {
                  concatVolume.setBaseObjectId(view.getId());
                  concatVolume.setBaseObjectType(ObjectType.pitView);
               }
            }
        }
    }

    private void augmentFlashCache(ObjectBundle bundle) {
        FlashCacheBundle fc = bundle.getFlashCacheBundle();
        if(fc == null){
            fc = new FlashCacheBundle();
            bundle.setFlashCacheBundle(fc);
        }

        FlashCache [] caches = fc.getFlashCache();
        if(caches == null){
            caches = new FlashCacheEx[0];
            fc.setFlashCache(caches);
        }

        FlashCacheMap[] cacheMap = fc.getFlashCacheMap();
        if(cacheMap == null){
            cacheMap = new FlashCacheMap[0];
            fc.setFlashCacheMap(cacheMap);
        }

        //Track the volumes that are being flashCached
        Set<String> allFlashCacheVolumes;
        if(cacheMap.length > 0){
            allFlashCacheVolumes = newHashSetWithExpectedSize(cacheMap.length);
        }
        else{
            allFlashCacheVolumes = Collections.emptySet();
        }

        final Drive [] drives = bundle.getDrive();

        //There should only be a maximum of one flashCache
        for(FlashCacheEx cache : (FlashCacheEx []) caches){
            final String cacheId = cache.getId();
            cache.setCachedVolumes(allFlashCacheVolumes);
            for(FlashCacheMap map : cacheMap){
                if(map.getFlashCache().equals(cache.getFlashCacheRef())){
                    allFlashCacheVolumes.add(map.getCachedVolume());
                }
            }

            //Determine which drives are associated with the flashCache
            Set<String> driveRefs = new HashSet<>();
            for(Drive d : drives){
                if(cacheId.equals(d.getCurrentVolumeGroupRef())){
                    driveRefs.add(d.getId());
                }
            }
            cache.setDriveRefs(driveRefs);
        }

        for(VolumeEx v : (VolumeEx[]) bundle.getVolume()){
            v.setFlashCached(allFlashCacheVolumes.contains(v.getId()));
        }

        for(ThinVolumeEx v : (ThinVolumeEx[]) bundle.getHighLevelVolBundle().getThinVolume()){
            v.setFlashCached(allFlashCacheVolumes.contains(v.getId()));
        }
    }

    private void augmentArvm(ObjectBundle bundle) {
        AsyncMirrorGroup [] rawGroups = bundle.getHighLevelVolBundle().getAsyncMirrorGroup();
        AsyncMirrorGroupMember [] rawMembers = bundle.getHighLevelVolBundle().getAsyncMirrorGroupMember();
        AsyncMirrorGroupIncompleteMember [] rawIncompleteMembers = bundle.getHighLevelVolBundle().getAsyncMirrorGroupIncompleteMember();
        if(rawGroups == null){
            bundle.getHighLevelVolBundle().setAsyncMirrorGroup(new Amg[0]);
            if(rawMembers == null){
                bundle.getHighLevelVolBundle().setAsyncMirrorGroupMember(new AmgMember[0]);
            }
            if(rawIncompleteMembers == null){
                bundle.getHighLevelVolBundle().setAsyncMirrorGroupIncompleteMember(new AmgIncompleteMember[0]);
            }
            return;
        }
        else if(rawMembers == null){
            rawMembers = new AsyncMirrorGroupMember[0];
            bundle.getHighLevelVolBundle().setAsyncMirrorGroupMember(rawMembers);
        }

        RemoteArray [] remoteArrays = bundle.getRemoteArray();
        if(remoteArrays == null){
            remoteArrays = new RemoteArray[0];
            bundle.setRemoteArray(remoteArrays);
        }

        Map<String, RemoteVolume> remoteVolumeMap;
        RemoteVolume [] remoteVolumes = bundle.getHighLevelVolBundle().getRemoteVol();
        if(remoteVolumes == null){
            bundle.getHighLevelVolBundle().setRemoteVol(new RemoteVolume[0]);
            remoteVolumeMap = Collections.emptyMap();
        }
        else if(remoteVolumes.length > 0){
            remoteVolumeMap = new HashMap<>(remoteVolumes.length);
            for(RemoteVolume volume : remoteVolumes){
                remoteVolumeMap.put(volume.getRemoteVolRef(), volume);
            }
        }
        else{
            remoteVolumeMap = Collections.emptyMap();
        }

        Map<String, RemoteTarget> remoteTargetMap;
        RemoteTarget [] targets = bundle.getStoragePoolBundle().getRemoteTarget();
        if(targets != null && targets.length > 0){
            remoteTargetMap = new HashMap<>(targets.length);
            for(RemoteTarget target : targets){
                remoteTargetMap.put(target.getRemoteRef(), target);
            }
        }
        else{
            remoteTargetMap = Collections.emptyMap();
        }
        Map<String, Amg> amgMap;
        Amg [] amgs = (Amg[]) rawGroups;
        if(amgs.length > 0) {
            amgMap = new HashMap<>(rawGroups.length);
            for (Amg amg : amgs) {
                amgMap.put(amg.getId(), amg);
                RemoteTarget target = remoteTargetMap.get(amg.getMirrorChannelRemoteTarget());
                if (target != null) {
                    amg.setConnectionType(target.getSCSIInitiatorTargetBaseProperties().getIoInterfaceType());
                    amg.setRemoteTarget(target);
                }
                for(RemoteArray remoteArray : remoteArrays){
                    for(String remoteTargetRef : remoteArray.getRemoteTargets()){
                        if(amg.getMirrorChannelRemoteTarget().equals(remoteTargetRef)){
                            amg.setRemoteTargetWwn(remoteArray.getSaId().getWorldWideName());
                            break;
                        }
                    }
                }
            }
            if(rawMembers.length > 0){
                AmgMember [] members = (AmgMember []) rawMembers;
                AmgIncompleteMember [] incompleteMembers = (AmgIncompleteMember[]) rawIncompleteMembers;
                VolumeEx [] volumes = (VolumeEx[]) bundle.getVolume();
                ThinVolumeEx [] thinVolumes = (ThinVolumeEx[]) bundle.getHighLevelVolBundle().getThinVolume();
                Map<String, MappableObject> volumeMap = new HashMap<>(volumes.length + thinVolumes.length);
                for(VolumeEx volume : volumes){
                    volumeMap.put(volume.getId(), volume);
                }
                for(ThinVolumeEx volume : thinVolumes){
                    volumeMap.put(volume.getId(), volume);
                }

                for(AmgIncompleteMember member : incompleteMembers){
                    Amg parent = amgMap.get(member.getGroupRef());
                    if(parent != null) {
                        member.setRemoteTargetWwn(parent.getRemoteTargetWwn());
                    }
                }

                for(AmgMember member : members){
                    RemoteVolume remoteVolume = remoteVolumeMap.get(member.getRemoteVolume());
                    if(remoteVolume != null){
                        member.setRemoteVolumeInfo(remoteVolume);
                        member.setRemoteVolumeName(remoteVolume.getRemoteVolUserLabel());
                    }
                    Amg parent = amgMap.get(member.getMirrorGroup());
                    if(parent != null) {
                        member.setRemoteTargetWwn(parent.getRemoteTargetWwn());
                    }
                    MappableObject localVolume = volumeMap.get(member.getLocalVolume());
                    if(localVolume != null){
                        member.setLocalVolumeName(localVolume.getName());
                    }
                    ConcatVolume concatVolume = getConcatVolume(member.getRepositoryVolume(), bundle);
                    if(concatVolume != null) {
                        member.setMediaScan(concatVolume.getMediaScanParams());
                        member.setTotalSizeInBytes(concatVolume.getAggregateCapacity());
                    }
                }
            }
        }
    }

    private void augmentTrays(ObjectBundle bundle) {
        TrayEx[] trays = (TrayEx[])bundle.getTray();
        TrayPosition[] trayPositions = bundle.getSa().getTrayPositionList().getTrayPosition();
        if (trayPositions.length > 0) {
            for (int i = 0; i < trayPositions.length; i++) {
                int trayId = trayPositions[i].getTrayId();
                for (TrayEx tray : trays) {
                    if (tray.getTrayId() == trayId) {
                        tray.trayPositionIndex = i;
                        break;
                    }
                }
            }
        }
    }

    private void augmentSaData(ObjectBundle obj) {
        SAData saData = obj.getSa().getSaData();
        saData.setFwVersion(ByteUtils.handleVersionString(saData.getFwVersion()));
        saData.setAppVersion(ByteUtils.handleVersionString(saData.getAppVersion()));
        saData.setBootVersion(ByteUtils.handleVersionString(saData.getBootVersion()));

        final StagedFirmware stagedFirmware = obj.getSa().getStagedFirmware();
        if(stagedFirmware != null) {
            stagedFirmware.setFwVersion(ByteUtils.handleVersionString(stagedFirmware.getFwVersion()));
            stagedFirmware.setAppVersion(ByteUtils.handleVersionString(stagedFirmware.getAppVersion()));
            stagedFirmware.setBootVersion(ByteUtils.handleVersionString(stagedFirmware.getBootVersion()));
        }
    }

    private Volume getThinVolumeRepository(ThinVolume v, ObjectBundle bundle){
        Volume ret = null;
        for(ConcatVolumeMember m : bundle.getHighLevelVolBundle().getConcatVolMember()){
            if(m.getConcatVolRef().equals(v.getRepositoryRef())){
                for(Volume vol : bundle.getVolume()){
                    if(vol.getVolumeRef().equals(m.getStorageVolumeRef())){
                        ret = vol;
                    }
                }
            }
        }
        return ret;
    }

    /**
     * Get the capacity in bytes of a ConcatVolume
     * @param repoId
     * @param bundle
     * @return
     */
    private long getConcatVolumeCapacity(String repoId, ObjectBundle bundle){
        long ret = 0;
        for(ConcatVolumeMember m : bundle.getHighLevelVolBundle().getConcatVolMember()){
            if(m.getConcatVolRef().equals(repoId)){
                for(Volume vol : bundle.getVolume()){
                    if(vol.getVolumeRef().equals(m.getStorageVolumeRef())){
                        ret += vol.getCapacity();
                    }
                }
            }
        }
        return ret;
    }

    private List<ConcatVolumeMember> getConcatVolumeMembers(String concatVolRef, ObjectBundle bundle){
        List<ConcatVolumeMember> ret = new ArrayList<>();
        for(ConcatVolumeMember m : bundle.getHighLevelVolBundle().getConcatVolMember()){
            if(m.getConcatVolRef().equals(concatVolRef)){
                ret.add(m);
            }
        }
        return ret;
    }

    private ConcatVolume getConcatVolume(String repoId, ObjectBundle bundle){
        ConcatVolume ret = null;
        for(ConcatVolume m : bundle.getHighLevelVolBundle().getConcatVolume()){
            if(m.getConcatVolRef().equals(repoId)){
                ret = m;
                break;
            }
        }
        return ret;
    }

    private void augmentPits(ObjectBundle graph){
        PIT [] tmp = graph.getHighLevelVolBundle().getPit();
        // for pre-fremont compatibility
        if(tmp == null){
            graph.getHighLevelVolBundle().setPit( new Snapshot [0] );
            return;
        }
        SnapshotGroup [] groups = (SnapshotGroup []) graph.getHighLevelVolBundle().getPitGroup();
        Snapshot [] pits = (Snapshot []) tmp;
        for(Snapshot snap : pits){
            for(SnapshotGroup group : groups){
                if(group.getId().equals(snap.getPitGroupRef())){
                    snap.setBaseVol(group.getBaseVolume());

                    if(group.isConsistencyGroup()) {
                        snap.setConsistencyGroupId(group.getConsistencyGroupRef());
                    }
                    break;
                }
            }
        }
    }

    private void augmentPitGroups(ObjectBundle graph){
        SnapshotGroup [] groups = (SnapshotGroup []) graph.getHighLevelVolBundle().getPitGroup();
        //For pre-fremont support
        if(groups == null){
            groups = new SnapshotGroup[0];
            graph.getHighLevelVolBundle().setPitGroup(groups);

        }
        PITConsistencyGroup [] cgroups = graph.getHighLevelVolBundle().getPitConsistencyGroup();
        if(cgroups == null){
            cgroups = new PITConsistencyGroup[0];
        }
        VolumeEx [] volumes = (VolumeEx []) graph.getVolume();
        VolumeCopy [] volumeCopies = graph.getHighLevelVolBundle().getVolumeCopy();

        Snapshot [] snapshots = (Snapshot[]) graph.getHighLevelVolBundle().getPit();
        Map<String, Integer> snapshotCounter = newHashMapWithExpectedSize(groups.length);
        if(snapshots == null){
            snapshots = new Snapshot[0];
            graph.getHighLevelVolBundle().setPit(snapshots);
        }
        else{
            for(Snapshot snap : snapshots){
                final String groupRef = snap.getPitGroupRef();
                snapshotCounter.merge(groupRef, 1, (a, b) -> a + b);
            }
        }

        for(SnapshotGroup g : groups){
            Integer count = snapshotCounter.get(g.getId());
            g.setSnapshotCount(count != null ? count : 0);

            ConcatVolume repository = getConcatVolume(g.getRepositoryVolume(), graph);
            if(repository != null){
                g.setRepositoryCapacity(repository.getAggregateCapacity());
            }
            else{
                continue;
            }

            //Define the label for any Snapshot group participating in a consistency group
            if( g.isConsistencyGroup() ){
                VolumeEx baseVolume = null;
                for(VolumeEx v : volumes){
                    if(v.getVolumeRef().equals(g.getBaseVolume())){
                        baseVolume = v;
                        break;
                    }
                }

                if(baseVolume == null){
                    continue;
                }

                for(PITConsistencyGroup cg : cgroups){
                    if(cg.getCgRef().equals(g.getConsistencyGroupRef())){
                        g.setLabel(cg.getLabel() + ":" + baseVolume.getLabel());
                        break;
                    }
                }
            }
            
            for (VolumeCopy vc : volumeCopies) {
                if(vc.getPgRef().equals(g.getId())) {
                    g.setVolcopyId(vc.getVolcopyRef());
                    break;
                }
            }
        }
    }

    private void augmentLegacySnapshots(ObjectBundle graph){
        SnapshotVolume [] tmp = graph.getHighLevelVolBundle().getSnapshotVol();
        if(tmp == null){
            return;
        }
        VolumeEx [] volumes = (VolumeEx [])graph.getVolume();

        LUNMapping[] mappings = graph.getStoragePoolBundle().getLunMapping();

        LegacySnapshotEx [] snaps = (LegacySnapshotEx []) tmp;
        for(LegacySnapshotEx v : snaps){
            determineVolumeUsage(v.getSnapshotRef(), v, graph);
            for(LUNMapping mapping : mappings){
                if(mapping.getVolumeRef().equals(v.getId())){
                    v.setMapping(mapping);
                }
            }
            v.setTotalSizeInBytes(v.getExtentSize());
            if(volumes != null){
                for(VolumeEx vol : volumes){
                    boolean repoFound = false, baseFound = false;

                    if(v.getRepositoryVolume().equals(vol.getId())){
                        v.setRepositoryCapacity(vol.getCapacity());
                        repoFound = true;
                    }

                    if(v.getBaseVolume().equals(vol.getId())){
                        v.setProtectionType(vol.getProtectionType());
                        v.setBaseVolumeSizeInBytes(vol.getCapacity());
                        v.setPreferredControllerId(vol.getPreferredControllerId());
                        baseFound = true;
                    }

                    if(repoFound && baseFound){
                        break;
                    }
                }
            }
        }
    }
    private void augmentThinVolumes(ObjectBundle graph){
        ThinVolume[] tmp = graph.getHighLevelVolBundle().getThinVolume();
        //for pre freemont
        if (tmp == null){
            graph.getHighLevelVolBundle().setThinVolume(new ThinVolumeEx[0]);
            return;
        }

        ThinVolumeEx [] thinvolumes = (ThinVolumeEx[]) tmp;
        LUNMapping[] mappings = graph.getStoragePoolBundle().getLunMapping();

        for(ThinVolumeEx v : thinvolumes){
            determineVolumeUsage(v.getVolumeRef(), v, graph);
            Volume repo = getThinVolumeRepository(v, graph);
            if(repo != null){
                v.setVolumeGroupRef(repo.getVolumeGroupRef());
                v.setStorageVolumeRef(repo.getVolumeRef());
                v.setBlkSize(repo.getBlkSize());
                v.setMediaScan(repo.getMediaScan());
            }
            if(v.getStatus().equals(ThinVolumeStatus.full)) {
                v.setVolumeFull(true);
            }
            else {
                v.setVolumeFull(false);
            }

            v.setDataAssurance(ProtectionType.type0Protection != v.getProtectionType());

            for(LUNMapping mapping : mappings){
                if(mapping.getVolumeRef().equals(v.getId())){
                    v.setMapping(mapping);
                }
            }

        }
    }

    private void augmentPitViews(ObjectBundle graph){
        PitViewEx [] views = (PitViewEx[]) graph.getHighLevelVolBundle().getPitView();
        //for pre fremont
        if (views == null){
            graph.getHighLevelVolBundle().setPitView(new PitViewEx[0]);
            return;
        }

        VolumeEx [] volumes = (VolumeEx []) graph.getVolume();
        ThinVolumeEx [] thinVolumes = (ThinVolumeEx[]) graph.getHighLevelVolBundle().getThinVolume();
        Snapshot [] pits = (Snapshot []) graph.getHighLevelVolBundle().getPit();
        LUNMapping[] mappings = graph.getStoragePoolBundle().getLunMapping();
        PITConsistencyGroupView [] groups = graph.getHighLevelVolBundle().getPitConsistencyGroupView();
        if(groups == null){
            groups = new PITConsistencyGroupView[0];
        }

        //Find the baseVolume
        for(PitViewEx v : views){

            determineVolumeUsage(v.getViewRef(), v, graph);

            if(pits != null){
                for(Snapshot p : pits){
                    if(p.getId().equals(v.getBasePIT())){
                        v.setTotalSizeInBytes(p.getPitCapacity());
                        break;
                    }
                }
            }

            //Thin volumes and standard volumes are valid bases
            CopyableObject baseVolume = null;
            for(VolumeEx bv : volumes){
                if(bv.getId().equals(v.getBaseVol())){
                    baseVolume = bv;
                    break;
                }
            }
            if(baseVolume == null && thinVolumes != null){
                for(ThinVolumeEx tv : thinVolumes){
                    if(tv.getId().equals(v.getBaseVol())){
                        baseVolume = tv;
                        break;
                    }
                }
            }
            //A valid baseVolume is required to continue
            if(baseVolume == null){
                continue;
            }

            v.setBaseVolumeCapacity(baseVolume.getTotalSizeInBytes());
            v.setProtectionType(baseVolume.getProtectionType());
            //Correctly identify the volumeCopy status of baseVolumes
            if(v.isVolumeCopySource()){
                baseVolume.setVolumeCopySource(true);
                if(v.isOnlineVolumeCopy()){
                    baseVolume.setOnlineVolumeCopy(true);
                }
            }

            if(v.getMembership().getViewType().equals(PITViewType.member)){
                String cgRef = v.getMembership().getCgViewRef();
                for(PITConsistencyGroupView g : groups){
                    if(g.getCgViewRef().equals(cgRef)){
                        // Define the consistency group id for PitViews belonging to a consistency group
                        v.setConsistencyGroupId(g.getGroupRef());

                        // Define the label for PitViews belonging to a consistency group
                        v.setLabel(g.getLabel() + ":" + baseVolume.getName());
                        break;
                    }
                }
            }

            if(v.getStatus().equals(PITViewStatus.full)){
                v.setVolumeFull(true);
            }
            else{
                v.setVolumeFull(false);
            }
            for(LUNMapping mapping : mappings){
                if(mapping.getVolumeRef().equals(v.getId())){
                    v.setMapping(mapping);
                }
            }

            long capacity = getConcatVolumeCapacity(v.getRepositoryVolume(), graph);
            v.setRepositoryCapacity(capacity);


        }
    }

    private void augmentVolumeGroups(ObjectBundle graph){
        VolumeGroupEx[] groups=(VolumeGroupEx[])graph.getVolumeGroup();
        Map<String,List<FreeExtent>> extents=new HashMap<String,List<FreeExtent>>();
        for (FreeExtent x:graph.getFreeExtent()){
            String vgRef=x.getVolumeGroupRef();
            if (vgRef != null){
                final List<FreeExtent> freeExtents = extents.computeIfAbsent(vgRef, k -> new ArrayList<>());
                freeExtents.add(x);
            }
        }
        for (VolumeGroupEx x:groups){

            DriveSpindleSpeed spindleSpeed = DriveSpindleSpeed.normalize(x.getSpindleSpeed());
            x.setNormalizedSpindleSpeed(spindleSpeed);

            long useSpace=calcUsedSpace(graph,x);
            long total=calcDPUsableSpace(graph, x);
            x.setUsedSpace(useSpace);
            x.setTotalRaidedSpace(total);
            x.setFreeSpace(total - useSpace);
            List<FreeExtent> freeExtents = extents.get(x.getId());
            if (freeExtents != null){
                x.setExtents(freeExtents);
                long max=0;
                for (FreeExtent y:freeExtents){
                    max=y.getRawCapacity() > max ? y.getRawCapacity() : max;
                }
                x.setLargestFreeExtentSize(max);
            }
            for (Volume v:graph.getVolume()){
                if (x.getVolumeGroupRef().equals(v.getVolumeGroupRef())){
                    VolumeStatus vStatus = v.getStatus();
                    if (vStatus.value > x.getRaidStatus().value){
                        x.setRaidStatus(vStatus);
                    }
                }
            }
            for(Drive d : graph.getDrive()){
                if(d.getCurrentVolumeGroupRef().equals(x.getVolumeGroupRef())){
                    x.setDriveMediaType(d.getDriveMediaType());
                    x.setDrivePhysicalType(d.getPhyDriveType());
                    break;
                }
            }
        }
    }

    /**
     * Retrieve a map of volumeRefs to values representing the amount of increase.
     * @param graph A non-null ObjectBundle
     * @return A Map of volume refs to their current expanding capacity, or an immutable empty map
     */
    private Map<String, Long> getVolumeExpansionData(final ObjectBundle graph){
        final DVEVolume [] expansionVolumes = graph.getDveVol();
        final Map<String, Long> ret;
        if(expansionVolumes == null){
            graph.setDveVol(new DVEVolume[0]);
            ret = Collections.emptyMap();
        }
        else if(expansionVolumes.length > 0){
            ret = newHashMapWithExpectedSize(expansionVolumes.length);
            for(DVEVolume v : expansionVolumes){
                ret.put(v.getVolume(), v.getAmountOfIncrease());
            }
        }
        else{
            ret = Collections.emptyMap();
        }
        return ret;
    }


    private void augmentVolume(ObjectBundle graph){
        Volume[] tmp = graph.getVolume();
        if (!(tmp instanceof VolumeEx[])){
            debugLogger.info("Un Augmented volume");
        }
        VolumeEx[] volumes = (VolumeEx[]) tmp;
        VolumeGroupEx [] volumeGroups = (VolumeGroupEx[]) graph.getVolumeGroup();
        Map<String, VolumeGroupEx> volumeGroupMap = newHashMapWithExpectedSize(volumeGroups.length);
        for(VolumeGroupEx vg : volumeGroups){
            volumeGroupMap.put(vg.getId(), vg);
        }

        Map<String, Long> expansionMap = getVolumeExpansionData(graph);

        volumeGroupMap = Collections.unmodifiableMap(volumeGroupMap);
        for (VolumeEx v: volumes){
            applyVolumeType(v.getVolumeRef(), v, graph);
            determineVolumeUsage(v.getVolumeRef(), v, graph);
			if (v.isAsyncMirrorSource ()) {
				v.setVolumeUse (VolumeUse.asyncMirrorSourceVolume);
			} else if (v.isAsyncMirrorTarget ()) {
				v.setVolumeUse (VolumeUse.asyncMirrorTargetVolume);
			}

            //Apply the data retrieved from DVEVolumes.
            final Long increasingBy = expansionMap.get(v.getId());
            if(increasingBy != null){
                v.setIncreasingBy(increasingBy);
            }

            v.setDataAssurance(ProtectionType.type0Protection != v.getProtectionType());

            VolumeGroupEx volumeGroup = volumeGroupMap.get(v.getVolumeGroupRef());
            if(volumeGroup != null){
                v.setDiskPool(RAIDLevel.raidDiskPool.equals(volumeGroup.getRaidLevel()));
            }
        }
        applyVolumeMappings(volumes,graph);
        AccessVolumeEx accessVolume = (AccessVolumeEx) graph.getSa().getAccessVolume();
        if (accessVolume != null){
            MappableObject[] x=new MappableObject[1];
            x[0]=accessVolume;
            applyVolumeMappings(x,graph);
        }
    }
    static private void applyVolumeMappings(MappableObject[] vols,ObjectBundle graph){

        LUNMapping[] mappings = graph.getStoragePoolBundle().getLunMapping();

        for(MappableObject v: vols) {
            List<LUNMapping> mappingsToUse=new ArrayList<>();
            String ref = v.getId();
            for (LUNMapping mapping : mappings) {
                String mapref = mapping.getVolumeRef();
                if (mapref.equals(ref)) {
                    mappingsToUse.add(mapping);
                }
            }
            v.setMappings(mappingsToUse);
        }

    }

    private static void determineVolumeUsage(String volRef, CopyableObject v, ObjectBundle graph)
    {
        final VolumeCopy[] volumeCopies = graph.getHighLevelVolBundle().getVolumeCopy();
        final AsyncMirrorGroupMember[] amgMembers = graph.getHighLevelVolBundle().getAsyncMirrorGroupMember();
        final AsyncMirrorGroup [] amGroups = graph.getHighLevelVolBundle().getAsyncMirrorGroup();
        final PITGroup[] pitGroups = graph.getHighLevelVolBundle().getPitGroup();
        final MirrorProxyVolume[] mirrorVolumes = graph.getHighLevelVolBundle().getMirrorVol();

        if(mirrorVolumes != null && mirrorVolumes.length > 0) {
            for (int i = 0, length = mirrorVolumes.length; i < length; i++) {
                MirrorProxyVolume proxy = mirrorVolumes[i];
                if(proxy.getBaseVolume().equals(volRef)){
                    if(MirrorProxyRole.primary.equals(proxy.getLocalRole())){
                        v.setRemoteMirrorSource(true);
                    }
                    else if(MirrorProxyRole.secondary.equals(proxy.getLocalRole())){
                        v.setRemoteMirrorTarget(true);
                    }
                }
            }
        }

        Map<String, AsyncMirrorGroup> amGroupLookup = new HashMap<>();
        if(amGroups != null){
            for(AsyncMirrorGroup grp : amGroups){
                amGroupLookup.put(grp.getGroupRef(), grp);
            }
        }
        if(amgMembers != null){
            for(AsyncMirrorGroupMember mg : amgMembers){
                if(mg.getLocalVolume().equals(volRef)){
                    AsyncMirrorGroup grp = amGroupLookup.get(mg.getMirrorGroup());
                    if(grp != null){
						v.setAsyncMirrorSource (AsyncMirrorGroupRole.primary == grp.getLocalRole ());
						v.setAsyncMirrorTarget (AsyncMirrorGroupRole.secondary == grp.getLocalRole ());
                    }
				}
            }
        }

        boolean isSource = false;
        boolean isTarget = false;
        boolean isCloneCopy = false;
        if(volumeCopies != null){
            for (VolumeCopy vc : volumeCopies)
            {
                String tgtRef = vc.getTargetVolume();
                String srcRef = vc.getSourceVolume();
                if (tgtRef.equals(volRef))
                {
                    isTarget = true;
                    isCloneCopy = vc.isCloneCopy();
                }
                else if(srcRef.equals(volRef)){
                    isSource = true;
                    isCloneCopy = vc.isCloneCopy();
                }
            }
        }
        v.setVolumeCopyTarget(isTarget);
        v.setVolumeCopySource(isSource);
        v.setOnlineVolumeCopy(isCloneCopy);

        boolean isPitBase = false;
        if (pitGroups != null)
        {
            for (PITGroup pg : pitGroups)
            {
                String baseRef = pg.getBaseVolume();

                if (baseRef.equals(volRef))
                {
                    isPitBase = true;
                    break;
                }
            }
        }
        v.setPitBaseVolume(isPitBase);
    }

    private static void applyVolumeType(String volRef, VolumeEx v, ObjectBundle graph)
    {
        final SnapshotVolume[] snaps = graph.getHighLevelVolBundle().getSnapshotVol();
        final MetadataVolume[] metaVolumes = graph.getHighLevelVolBundle().getMetadataVol();
        final MirrorProxyVolume[] mirrorVolumes = graph.getHighLevelVolBundle().getMirrorVol();
        final VolumeCopy[] volumeCopies = graph.getHighLevelVolBundle().getVolumeCopy();
//        final AsyncMirrorGroupMember[] amgMembers = graph.getHighLevelVolBundle().getAsyncMirrorGroupMember();
//        final AsyncMirrorGroupIncompleteMember[] amgIncompleteMembers = graph.getHighLevelVolBundle()
//                .getAsyncMirrorGroupIncompleteMember();
        final PITGroup[] pitGroups = graph.getHighLevelVolBundle().getPitGroup();

        //set the default
        v.setVolumeUse(VolumeUse.standardVolume);
        if (isValidRepositoryName(v.getName()))
        {
            v.setVolumeUse(VolumeUse.freeRepositoryVolume);
        }

        final ConcatVolumeMember[] concatVolMember = graph.getHighLevelVolBundle().getConcatVolMember();
        if (concatVolMember != null)
        {
            for (ConcatVolumeMember cv : concatVolMember)
            {
                String catRef = cv.getStorageVolumeRef();
                if (volRef.equals(catRef))
                {
                    v.setVolumeUse(VolumeUse.concatVolume);
                    break;
                }
            }
        }

        for (SnapshotVolume snap : snaps)
        {
            String repoRef = snap.getRepositoryVolume();

            if (repoRef.equals(volRef))
            {
                v.setVolumeUse(VolumeUse.repositoryVolume);
                v.setVolumeFull(snap.isRepositoryFull());
                break;
            }
        }

        for (MetadataVolume mv : metaVolumes)
        {
            String repoRef = mv.getAssociatedVolume();
            if (repoRef.equals(volRef))
            {
                v.setVolumeUse(VolumeUse.repositoryVolume);
                break;
            }
        }


        boolean isTarget = false;
        for (VolumeCopy vc : volumeCopies)
        {
            String targetref = vc.getTargetVolume();
            if (targetref.equals(volRef))
            {
                isTarget = true;
                break;
            }
        }
        v.setVolumeCopyTarget(isTarget);

        boolean isPitBase = false;
        if (pitGroups != null)
        {
            for (PITGroup pg : pitGroups)
            {
                String baseRef = pg.getBaseVolume();

                if (baseRef.equals(volRef))
                {
                    isPitBase = true;
                    break;
                }
            }
        }
        v.setPitBaseVolume(isPitBase);

        if(mirrorVolumes != null && mirrorVolumes.length > 0) {
            for (int i = 0, length = mirrorVolumes.length; i < length; i++) {
                MirrorProxyVolume proxy = mirrorVolumes[i];
                if(proxy.getBaseVolume().equals(volRef)){
                    if(MirrorProxyRole.primary.equals(proxy.getLocalRole())){
                        v.setVolumeUse(VolumeUse.mirrorSourceVolume);
                    }
                    else if(MirrorProxyRole.secondary.equals(proxy.getLocalRole())){
                        v.setVolumeUse(VolumeUse.mirrorTargetVolume);
                    }
                }
            }
        }

    }


    private void augmentHostGroup(ObjectBundle obj){
    }

    private void augmentInterface(ObjectBundle obj){
        for (Controller c:obj.getController()){
            IOInterfaceTypeData[] ioi = c.getHostInterfaces();
            for (IOInterfaceTypeData iotd:ioi){
                IOInterfaceType type = iotd.getInterfaceType();
                HostInterfacePort port;
                switch (type){
                    case fc:
                        port=(HostInterfacePort) iotd.getFibre();
                        break;
                    case iscsi:
                        IScsiInterfacePort iscsi = (IScsiInterfacePort) iotd.getIscsi();
                        port= iscsi;
                        Target[] targets = obj.getStoragePoolBundle().getTarget();
                        for (Target target : targets) {
                            if (target.getNodeName().getIoInterfaceType() == IOInterfaceType.iscsi) {
                                ScsiNodeName nodeName = target.getNodeName();
                                String iqn = nodeName.getIscsiNodeName();
                                iscsi.setIqn(iqn);
                            }
                        }

                        break;
                    case sas:
                        port=(HostInterfacePort) iotd.getSas();
                        break;
                    case ib:
                        port=(HostInterfacePort) iotd.getIb();
                        break;
                    default:{
                        port=null;
                    }
                }
                if (port !=null){
                    port.setControllerId(c.getControllerRef());
                }
            }
        }
    }
    private void augmentHosts(ObjectBundle graph){
        HostEx[] ret =(HostEx[]) graph.getStoragePoolBundle().getHost();

        for (HostEx host:ret){
            List<HostPort> ports = new ArrayList<>();
            HostPort[] hostPort = graph.getStoragePoolBundle().getHostPort();
            List<HostSidePort> hostSide=new ArrayList<>();
            for(HostPort port : hostPort){
                if(port.getHostRef().equals(host.getHostRef())){
                    ports.add(port);
                    HostSidePort x=new HostSidePort();
                    SCSIInitiatorPortId portId = port.getPortId();
                    if (portId != null) {
                        IOInterfaceType ioInterfaceType = portId.getIoInterfaceType();
                        x.setType(ioInterfaceType);
                        x.setLabel(port.getLabel());
                        switch (ioInterfaceType) {
                            case ib: {
                                x.setAddress(portId.getIbPortName());
                                break;
                            }
                            case iscsi: {
                                //in practice this does not work.  ISCSO is handled as initiators
                                x.setAddress(portId.getIscsiPortName().getInitiatorNodeName());
                                break;
                            }
                            case sas: {
                                x.setAddress(portId.getSasPortName());
                                break;
                            }
                            default: {
                                String portName = portId.getPortName();
                                portName = portName == null ? "" : portName;
                                x.setAddress(portName);
                                break;

                            }
                        }
                    }
                    hostSide.add(x);
                }
            }
            List<Initiator> initiators = new ArrayList<>();
            Initiator[] initiator1 = graph.getStoragePoolBundle().getInitiator();
            if (initiator1 != null) {
                for (Initiator initiator : initiator1) {
                    if (initiator.getHostRef().equals(host.getHostRef())) {
                        String nodeName = null;
                        ScsiNodeName node = initiator.getNodeName();
                        IOInterfaceType type = node.getIoInterfaceType();
                        switch (type) {
                            case iscsi:
                                nodeName = node.getIscsiNodeName();
                                break;
                            case nvmeof:
                                nodeName = node.getNvmeNodeName();
                                break;
                            default:
                                debugLogger.warn("Unknown interface type for host: {}", type);
                                break;
                        }

                        initiators.add(initiator);
                        HostSidePort x = new HostSidePort(type, nodeName, initiator.getLabel());
                        hostSide.add(x);
                    }
                }
            }
            host.setInitiators(initiators);
            host.setPorts(ports);
            host.setHostSidePorts(hostSide);
        }

    }

    private void augmentHostTypeValues(ObjectBundle graph)
    {
        HostTypeValues[] ret = (HostTypeValues[]) graph.getSa().getHostSpecificVals();
        int defaultIndex = graph.getSa().getDefaultHostTypeIndex();
        for (int i = 0; i < ret.length; i++)
        {
            ret[i].setIndex(i);
            ret[i].setDefault(i == defaultIndex);
        }
    }

    private void augmentDrive(ObjectBundle graph)
    {
        DriveEx[] drives = (DriveEx[]) graph.getDrive();
        Interposer[] interposers = graph.getInterposer();
        LockKeyID[] lockkeyIDs = graph.getLockKeyIDs();
        for(DriveEx d : drives)
        {
            if(d.isInterposerPresent())
            {
                //If interposer is present, get the firmware version from the interposer
                String interposerRef = d.getInterposerRef();
                d.setFirmwareVersion(getDriveFirmwareVersion(interposerRef, interposers));
            }
            else
            {
                d.setFirmwareVersion(d.getSoftwareVersion());
            }

            d.setLockKeyIDValue(getLockKeyID(d.getLockKeyID(), lockkeyIDs));
        }
    }


    private void augmentBattery(ObjectBundle graph)
    {
        BatteryEx[] batteries = (BatteryEx[]) graph.getComponentBundle().getBattery();
        for(BatteryEx b : batteries)
        {
            if(graph.getSa().getFeatureParameters().isSmartBatteryTechnology())
            {
                //If smart battery technology is supported set smart battery data
                b.setSmartBatteryData(graph.getSa().getSmartBatteryData());
            }
            else {
                b.setSmartBatteryData(null);
            }
        }
    }

    private String getDriveFirmwareVersion(String interposerRef, Interposer[] interposers)
    {
        String version = null;
        if(interposers != null) {
            for (Interposer i : interposers) {
                if (i.getInterposerRef().equals(interposerRef)) {
                    InterposerData iData = i.getInterposerTypeData().getFcSasInterposer();
                    version = iData.getDriveFwVersion();
                    break;
                }
            }
        }
        return version;
    }

    private String getLockKeyID(String driveLockKeyIDRef, LockKeyID[] lockKeyIDs)
    {
        String value = null;
        if(lockKeyIDs != null) {
            for (LockKeyID l : lockKeyIDs) {
                if (l.getLockKeyIDRef().equals(driveLockKeyIDRef)) {
                    value = l.getLockKeyID();
                    break;
                }
            }
        }
        return value;
    }


    /**
     * Determines if the name passed in is a valid repository name.
     * <P>
     * NOTES/ASSUMPTIONS:  <BR>
     * None.
     *
     * @param name to check if valid repository name.
     * @return true if name can be used as a repository volume name, false otherwise.
     */
    public static boolean isValidRepositoryName(String name)
    {
        return name.toLowerCase().matches(REPOS_REGEX);
    }

    static private long calcDPUsableSpace(ObjectBundle graph, VolumeGroup group)
    {
        long capacity = calcUsedSpace(graph, group);
        String groupRef = group.getVolumeGroupRef();

        for(FreeExtent fe : graph.getFreeExtent())
        {
            if (groupRef.equals(fe.getVolumeGroupRef()))
            {
                capacity += fe.getRawCapacity();
            }
        }
        return capacity;
    }

    static private long calcUsedSpace(ObjectBundle graph, VolumeGroup group)
    {
        long ret = 0;
        devmgr.api.symbol.Volume[] volumes = graph.getVolume();
        long allocGranularity = 0;
        if (group.getRaidLevel() == RAIDLevel.raidDiskPool && group.getVolumeGroupData().getType().equals(VolumeGroupType.diskPool))
        {
            allocGranularity = group.getVolumeGroupData().getDiskPoolData().getAllocGranularity();
        }
        for (int i = 0; i < volumes.length; i++)
        {
            devmgr.api.symbol.Volume v = volumes[i];
            String vgref = v.getVolumeGroupRef();
            if (vgref != null)
            {
                if (vgref.equals(group.getVolumeGroupRef()))
                {
                    long volumeCapacity = v.getCapacity();
                    if (allocGranularity != 0)
                    {
                        // Adjust volume capacity to the allocation granularity boundary
                        long modCapacity = volumeCapacity % allocGranularity;
                        if (modCapacity != 0)
                        {
                            volumeCapacity += allocGranularity - modCapacity;
                        }
                    }
                    ret += volumeCapacity;
                }
            }
        }
        return ret;
    }
}
